# tests.test_cluster.test_elbow
# Tests for the KElbowVisualizer
#
# Author:   Benjamin Bengfort
# Created:  Thu Mar 23 22:30:19 2017 -0400
#
# Copyright (C) 2017 The scikit-yb developers
# For license information, see LICENSE.txt
#
# ID: test_elbow.py [5a370c8] benjamin@bengfort.com $

"""
Tests for the KElbowVisualizer
"""

##########################################################################
## Imports
##########################################################################

import sys
import pytest
import numpy as np
import matplotlib.pyplot as plt

from scipy.sparse import csc_matrix, csr_matrix
from numpy.testing import assert_array_almost_equal

from sklearn.datasets import make_blobs
from sklearn.cluster import KMeans, MiniBatchKMeans
from sklearn.feature_extraction.text import TfidfVectorizer

from tests.fixtures import Dataset
from tests.base import VisualTestCase
from yellowbrick.datasets import load_hobbies
from yellowbrick.cluster.elbow import distortion_score
from yellowbrick.cluster.elbow import KElbowVisualizer, kelbow_visualizer
from yellowbrick.exceptions import YellowbrickValueError, YellowbrickWarning

from tests.base import IS_WINDOWS_OR_CONDA

try:
    import pandas as pd
except ImportError:
    pd = None

##########################################################################
## Data
##########################################################################


@pytest.fixture(scope="class")
def clusters(request):
    # TODO: replace with make_blobs
    X = np.array(
        [
            [-0.40020753, -4.67055317, -0.27191127, -1.49156318],
            [0.37143349, -4.89391622, -1.23893945, 0.48318165],
            [8.625142, -1.2372284, 1.39301471, 4.3394457],
            [7.65803596, -2.21017215, 1.99175714, 3.71004654],
            [0.89319875, -5.37152317, 1.50313598, 1.95284886],
            [2.68362166, -5.78810913, -0.41233406, 1.94638989],
            [7.63541182, -1.99606076, 0.9241231, 4.53478238],
            [9.04699415, -0.74540679, 0.98042851, 5.99569071],
            [1.02552122, -5.73874278, -1.74804915, -0.07831216],
            [7.18135665, -3.49473178, 1.14300963, 4.46065816],
            [0.58812902, -4.66559815, -0.72831685, 1.40171779],
            [1.48620862, -5.9963108, 0.19145963, -1.11369256],
            [7.6625556, -1.21328083, 2.06361094, 6.2643551],
            [9.45050727, -1.36536078, 1.31154384, 3.89103468],
            [6.88203724, -1.62040255, 3.89961049, 2.12865388],
            [5.60842705, -2.10693356, 1.93328514, 3.90825432],
            [2.35150936, -6.62836131, -1.84278374, 0.51540886],
            [1.17446451, -5.62506058, -2.18420699, 1.21385128],
        ]
    )

    y = np.array([0, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1, 1, 1, 1, 0, 0])

    request.cls.clusters = Dataset(X, y)


##########################################################################
## K-Elbow Helpers Test Cases
##########################################################################


@pytest.mark.usefixtures("clusters")
class TestKElbowHelper(object):
    """
    Helper functions for K-Elbow Visualizer
    """

    def test_distortion_score(self):
        """
        Test the distortion score metric function
        """
        score = distortion_score(self.clusters.X, self.clusters.y)
        assert score == pytest.approx(69.10006514142941)

    @pytest.mark.parametrize("func", [csc_matrix, csr_matrix], ids=["csc", "csr"])
    def test_distortion_score_sparse_matrix_input(self, func):
        """
        Test the distortion score metric on a sparse array
        """
        score = distortion_score(func(self.clusters.X), self.clusters.y)
        assert score == pytest.approx(69.10006514142938)

    @pytest.mark.skipif(pd is None, reason="pandas is required")
    def test_distortion_score_pandas_input(self):
        """
        Test the distortion score metric on pandas DataFrame and Series
        """
        df = pd.DataFrame(self.clusters.X)
        s = pd.Series(self.clusters.y)

        score = distortion_score(df, s)
        assert score == pytest.approx(69.10006514142941)

    def test_distortion_score_empty_clusters(self):
        """
        Ensure no ValueError is thrown when there are empty clusters #1185
        """
        X = np.array([[1, 2], [3, 4], [5, 6]])
        valuea = distortion_score(X, np.array([1, 3, 3]))
        valueb = distortion_score(X, np.array([0, 1, 1]))
        assert valuea == valueb


##########################################################################
## KElbowVisualizer Test Cases
##########################################################################


@pytest.mark.usefixtures("clusters")
class TestKElbowVisualizer(VisualTestCase):
    """
    K-Elbow Visualizer Tests
    """

    @pytest.mark.xfail(reason="images not close due to timing lines")
    @pytest.mark.filterwarnings("ignore:No 'knee'")
    def test_integrated_kmeans_elbow(self):
        """
        Test no exceptions for kmeans k-elbow visualizer on blobs dataset
        """
        # NOTE #182: cannot use occupancy dataset because of memory usage

        # Generate a blobs data set
        X, y = make_blobs(
            n_samples=1000, n_features=12, centers=6, shuffle=True, random_state=42
        )

        try:
            _, ax = plt.subplots()

            visualizer = KElbowVisualizer(KMeans(random_state=42), k=4, ax=ax)
            visualizer.fit(X)
            visualizer.finalize()

            self.assert_images_similar(visualizer)
        except Exception as e:
            pytest.fail("error during k-elbow: {}".format(e))

    @pytest.mark.xfail(reason="images not close due to timing lines")
    @pytest.mark.filterwarnings("ignore:No 'knee'")
    def test_integrated_mini_batch_kmeans_elbow(self):
        """
        Test no exceptions for mini-batch kmeans k-elbow visualizer
        """
        # NOTE #182: cannot use occupancy dataset because of memory usage

        # Generate a blobs data set
        X, y = make_blobs(
            n_samples=1000, n_features=12, centers=6, shuffle=True, random_state=42
        )

        try:
            _, ax = plt.subplots()

            visualizer = KElbowVisualizer(MiniBatchKMeans(random_state=42), k=4, ax=ax)
            visualizer.fit(X)
            visualizer.finalize()

            self.assert_images_similar(visualizer)
        except Exception as e:
            pytest.fail("error during k-elbow: {}".format(e))

    @pytest.mark.skip(reason="takes over 20 seconds to run")
    def test_topic_modeling_k_means(self):
        """
        Test topic modeling k-means on the hobbies corpus
        """
        corpus = load_hobbies()

        tfidf = TfidfVectorizer()
        docs = tfidf.fit_transform(corpus.data)
        visualizer = KElbowVisualizer(KMeans(), k=(4, 8))

        visualizer.fit(docs)
        visualizer.finalize()

        self.assert_images_similar(visualizer)

    def test_invalid_k(self):
        """
        Assert that invalid values of K raise exceptions
        """
        # Generate a blobs data set
        X, y = make_blobs(
            n_samples=1000, n_features=12, centers=6, shuffle=True, random_state=42
        )

        with pytest.raises(YellowbrickValueError):
            KElbowVisualizer(KMeans(), k=(1, 2, 3, "foo", 5)).fit(X)

        with pytest.raises(YellowbrickValueError):
            KElbowVisualizer(KMeans(), k="foo").fit(X)

    def test_valid_k(self):
        """
        Assert that valid values of K generate correct k_values_
        """
        # if k is an int, k_values_ = range(2, k+1)
        # if k is a tuple of 2 ints, k_values = range(k[0], k[1])
        # if k is an iterable, k_values_ = list(k)

        # Generate a blobs data set
        X, y = make_blobs(
            n_samples=1000, n_features=12, centers=6, shuffle=True, random_state=42
        )

        visualizer = KElbowVisualizer(KMeans(), k=8).fit(X)
        assert visualizer.k_values_ == list(np.arange(2, 8 + 1))

        visualizer = KElbowVisualizer(KMeans(), k=(4, 12)).fit(X)
        assert visualizer.k_values_ == list(np.arange(4, 12))

        visualizer = KElbowVisualizer(KMeans(), k=np.arange(10, 100, 10)).fit(X)
        assert visualizer.k_values_ == list(np.arange(10, 100, 10))

        visualizer = KElbowVisualizer(
            KMeans(), k=[10, 20, 30, 40, 50, 60, 70, 80, 90]
        ).fit(X)
        assert visualizer.k_values_ == list(np.arange(10, 100, 10))

    @pytest.mark.xfail(sys.platform == "win32", reason="images not close on windows")
    def test_distortion_metric(self):
        """
        Test the distortion metric of the k-elbow visualizer
        """
        visualizer = KElbowVisualizer(
            KMeans(random_state=0),
            k=5,
            metric="distortion",
            timings=False,
            locate_elbow=False,
        )
        visualizer.fit(self.clusters.X)

        expected = np.array([69.100065, 54.891057, 44.319888, 35.857462])
        assert len(visualizer.k_scores_) == 4

        visualizer.finalize()
        self.assert_images_similar(visualizer, tol=0.03)
        assert_array_almost_equal(visualizer.k_scores_, expected)

    @pytest.mark.xfail(sys.platform == "win32", reason="images not close on windows")
    def test_silhouette_metric(self):
        """
        Test the silhouette metric of the k-elbow visualizer
        """
        visualizer = KElbowVisualizer(
            KMeans(random_state=0),
            k=5,
            metric="silhouette",
            timings=False,
            locate_elbow=False,
        )
        visualizer.fit(self.clusters.X)

        expected = np.array([0.691636, 0.453478, 0.242102, 0.235422])
        assert len(visualizer.k_scores_) == 4

        visualizer.finalize()
        self.assert_images_similar(visualizer, tol=0.03)
        assert_array_almost_equal(visualizer.k_scores_, expected)

    @pytest.mark.xfail(sys.platform == "win32", reason="images not close on windows")
    def test_calinski_harabasz_metric(self):
        """
        Test the calinski-harabasz metric of the k-elbow visualizer
        """
        visualizer = KElbowVisualizer(
            KMeans(random_state=0),
            k=5,
            metric="calinski_harabasz",
            timings=False,
            locate_elbow=False,
        )
        visualizer.fit(self.clusters.X)
        assert len(visualizer.k_scores_) == 4
        assert visualizer.elbow_value_ is None

        expected = np.array([81.662726, 50.129783, 39.744834, 34.978841])

        visualizer.finalize()
        self.assert_images_similar(visualizer)
        assert_array_almost_equal(visualizer.k_scores_, expected)

    @pytest.mark.xfail(sys.platform == "win32", reason="images not close on windows")
    def test_distance_metric(self):
        """
        Test the manhattan distance metric of distortion for k-elbow
        """
        visualizer = KElbowVisualizer(
            KMeans(random_state=0),
            k=5,
            metric="distortion",
            distance_metric="manhattan",
            timings=False,
            locate_elbow=False,
        )
        visualizer.fit(self.clusters.X)
        assert len(visualizer.k_scores_) == 4
        assert visualizer.elbow_value_ is None

        expected = np.array([189.06013, 152.276395, 132.668674, 110.741248])

        visualizer.finalize()
        self.assert_images_similar(visualizer)
        assert_array_almost_equal(visualizer.k_scores_, expected)

    @pytest.mark.xfail(
        IS_WINDOWS_OR_CONDA,
        reason="font rendering different in OS and/or Python; see #892",
    )
    def test_locate_elbow(self):
        """
        Test the addition of locate_elbow to an image
        """
        X, y = make_blobs(
            n_samples=1000, n_features=5, centers=3, shuffle=True, random_state=42
        )

        visualizer = KElbowVisualizer(
            KMeans(random_state=0),
            k=6,
            metric="calinski_harabasz",
            timings=False,
            locate_elbow=True,
        )
        visualizer.fit(X)

        # Test the important assertions
        assert len(visualizer.k_scores_) == 5
        assert visualizer.elbow_value_ == 3

        # Test that scores are within 5% of expected values
        # (allows for BLAS implementation differences across platforms)
        expected = np.array([4286.5, 12463.4, 8766.8, 6950.1, 5863.6])
        relative_error = np.abs(visualizer.k_scores_ - expected) / expected

        assert np.all(relative_error < 0.05), (
            f"Calinski-Harabasz scores exceed 5% relative error: \n"
            f"Expected: {expected}\n"
            f"Actual: {visualizer.k_scores_}\n"
            f"Relative error: {relative_error * 100}%"
        )

        visualizer.finalize()
        self.assert_images_similar(visualizer, tol=0.5, windows_tol=2.2)

    def test_no_knee(self):
        """
        Assert that a warning is issued if there is no knee detected
        """
        X, y = make_blobs(n_samples=1000, centers=3, n_features=12, random_state=12)
        message = (
            "No 'knee' or 'elbow point' detected "
            "This could be due to bad clustering, no "
            "actual clusters being formed etc."
        )
        with pytest.warns(YellowbrickWarning, match=message):
            visualizer = KElbowVisualizer(
                KMeans(random_state=12), k=(4, 12), locate_elbow=True
            )
            visualizer.fit(X)

    def test_bad_metric(self):
        """
        Assert KElbow raises an exception when a bad metric is supplied
        """
        with pytest.raises(YellowbrickValueError):
            KElbowVisualizer(KMeans(), k=5, metric="foo")

    def test_bad_distance_metric(self):
        """
        Assert KElbow raises an exception when a bad distance metric is supplied
        """
        with pytest.raises(YellowbrickValueError):
            KElbowVisualizer(KMeans(), k=5, distance_metric="foo")

    @pytest.mark.xfail(
        IS_WINDOWS_OR_CONDA,
        reason="font rendering different in OS and/or Python; see #892",
    )
    def test_timings(self):
        """
        Test the twinx double axes with k-elbow timings
        """
        visualizer = KElbowVisualizer(
            KMeans(random_state=0), k=5, timings=True, locate_elbow=False
        )
        visualizer.fit(self.clusters.X)

        # Check that we kept track of time
        assert len(visualizer.k_timers_) == 4
        assert all([t > 0 for t in visualizer.k_timers_])

        # Check that we plotted time on a twinx
        assert hasattr(visualizer, "axes")
        assert len(visualizer.axes) == 2

        # delete the timings axes and
        # overwrite k_timers_, k_values_ for image similarity Tests
        visualizer.axes[1].remove()
        visualizer.k_timers_ = [
            0.01084589958190918,
            0.011144161224365234,
            0.017028093338012695,
            0.010634183883666992,
        ]
        visualizer.k_values_ = [2, 3, 4, 5]

        # call draw again which is normally called in fit
        visualizer.draw()
        visualizer.finalize()

        self.assert_images_similar(visualizer)

    def test_sample_weights(self):
        """
        Test that passing in sample weights correctly influences the clusterer's fit
        """
        seed = 1234

        # original data has 5 clusters
        X, y = make_blobs(
            n_samples=[5, 30, 30, 30, 30],
            n_features=5,
            random_state=seed,
            shuffle=False,
        )

        visualizer = KElbowVisualizer(
            KMeans(random_state=seed), k=(2, 12), timings=False
        )
        visualizer.fit(X)
        assert visualizer.elbow_value_ == 5

        # weights should push elbow down to 4
        weights = np.concatenate([np.ones(5) * 0.0001, np.ones(120)])

        visualizer.fit(X, sample_weight=weights)
        assert visualizer.elbow_value_ == 4

    @pytest.mark.xfail(reason="images not close due to timing lines")
    def test_quick_method(self):
        """
        Test the quick method producing a valid visualization
        """
        X, y = make_blobs(
            n_samples=1000, n_features=12, centers=8, shuffle=False, random_state=2
        )

        model = MiniBatchKMeans(3, random_state=43)
        oz = kelbow_visualizer(model, X, show=False)
        assert isinstance(oz, KElbowVisualizer)

        self.assert_images_similar(oz)

    def test_quick_method_params(self):
        """
        Test the quick method correctly consumes the user-provided parameters
        """
        X, y = make_blobs(centers=3)
        custom_title = "My custom title"
        model = KMeans(3, random_state=13)
        oz = kelbow_visualizer(
            model, X, sample_weight=np.ones(X.shape[0]), title=custom_title, show=False
        )
        assert oz.title == custom_title

    @pytest.mark.xfail(sys.platform == "win32", reason="images not close on windows")
    def test_set_colors_manually(self):
        """
        Test the silhouette metric of the k-elbow visualizer
        """
        oz = KElbowVisualizer(
            KMeans(random_state=0),
            k=5,
        )

        oz.metric_color = "r"
        oz.timing_color = "y"
        oz.vline_color = "c"

        # Create artificial "fit" data for testing purposes
        oz.k_values_ = [1, 2, 3, 4, 5, 6, 7, 8]
        oz.k_timers_ = [6.2, 8.3, 10.1, 15.8, 21.2, 27.9, 38.2, 44.9]
        oz.k_scores_ = [0.8, 0.7, 0.55, 0.48, 0.40, 0.38, 0.35, 0.30]
        oz.elbow_value_ = 5
        oz.elbow_score_ = 0.40

        # Execute drawing
        oz.draw()
        oz.finalize()
        self.assert_images_similar(oz, tol=3.2)

    def test_get_params(self):
        """
        Ensure the get params works for sklearn-compatibility
        """
        oz = KElbowVisualizer(
            KMeans(random_state=0),
            k=5,
        )
        params = oz.get_params()
        assert len(params) > 0



# tests.test_text.test_postag
# Tests for the part-of-speech tagging visualization
#
# Author:   Rebecca Bilbro
# Created:  2019-02-19 21:29
#
# Copyright (C) 2019 The scikit-yb developers
# For license information, see LICENSE.txt
#
# ID: test_postag.py [bd9cbb9] rebecca.bilbro@bytecubed.com $

"""
Tests for the part-of-speech tagging visualization
"""

##########################################################################
## Imports
##########################################################################

import pytest
import matplotlib.pyplot as plt

from tests.base import VisualTestCase
from tests.base import IS_WINDOWS_OR_CONDA

from yellowbrick.text.postag import *
from yellowbrick.exceptions import YellowbrickValueError

try:
    import nltk
    from nltk import pos_tag, sent_tokenize
    from nltk import word_tokenize, wordpunct_tokenize
except ImportError:
    nltk = None

try:
    import spacy
except ImportError:
    spacy = None


##########################################################################
## Data
##########################################################################

sonnets = [
    """
    FROM fairest creatures we desire increase,
    That thereby beauty's rose might never die,
    But as the riper should by time decease,
    His tender heir might bear his memory:
    But thou, contracted to thine own bright eyes,
    Feed'st thy light'st flame with self-substantial fuel,
    Making a famine where abundance lies,
    Thyself thy foe, to thy sweet self too cruel.
    Thou that art now the world's fresh ornament
    And only herald to the gaudy spring,
    Within thine own bud buriest thy content
    And, tender churl, makest waste in niggarding.
    Pity the world, or else this glutton be,
    To eat the world's due, by the grave and thee.
    """,
    """
    When forty winters shall beseige thy brow,
    And dig deep trenches in thy beauty's field,
    Thy youth's proud livery, so gazed on now,
    Will be a tatter'd weed, of small worth held:
    Then being ask'd where all thy beauty lies,
    Where all the treasure of thy lusty days,
    To say, within thine own deep-sunken eyes,
    Were an all-eating shame and thriftless praise.
    How much more praise deserved thy beauty's use,
    If thou couldst answer 'This fair child of mine
    Shall sum my count and make my old excuse,'
    Proving his beauty by succession thine!
    This were to be new made when thou art old,
    And see thy blood warm when thou feel'st it cold.
    """,
    """
    Look in thy glass, and tell the face thou viewest
    Now is the time that face should form another;
    Whose fresh repair if now thou not renewest,
    Thou dost beguile the world, unbless some mother.
    For where is she so fair whose unear'd womb
    Disdains the tillage of thy husbandry?
    Or who is he so fond will be the tomb
    Of his self-love, to stop posterity?
    Thou art thy mother's glass, and she in thee
    Calls back the lovely April of her prime:
    So thou through windows of thine age shall see
    Despite of wrinkles this thy golden time.
    But if thou live, remember'd not to be,
    Die single, and thine image dies with thee.
    """,
]


##########################################################################
## PosTag Utils
##########################################################################


def check_nltk_data():
    """
    Returns True if NLTK data has been downloaded, False otherwise
    """
    try:
        nltk.data.find("corpora/treebank")
        return True
    except LookupError:
        pytest.xfail("error occured because nltk postag data is not available")


def check_spacy_data():
    """
    Returns True if SpaCy data has been downloaded, False otherwise
    """
    try:
        spacy.load("en_core_web_sm")
        return True
    except OSError:
        pytest.xfail("error occured because spacy data model is not available")


def get_tagged_docs(X, model="nltk", tagger="word"):
    """
    X is a list of strings; each string is a single document.
    For each document, perform part-of-speech tagging, and
    yield a list of sentences, where each sentence is a list
    of (token, tag) tuples

    If model=="nltk", `NLTK` will be used to sentence and word
    tokenize the incoming documents. User may select the `NLTK`
    tagger to be used; (for now) either the word tokenizer or the
    workpunct tokenizer.

    If model=="spacy", `SpaCy` will be used to sentence and word
    tokenize the incoming documents.
    """
    if model == "spacy":
        nlp = spacy.load("en_core_web_sm")
        for doc in X:
            tagged = nlp(doc)
            yield [
                list((token.text, token.pos_) for token in sent)
                for sent in tagged.sents
            ]

    elif model == "nltk":
        if tagger == "wordpunct":
            for doc in X:
                yield [pos_tag(wordpunct_tokenize(sent)) for sent in sent_tokenize(doc)]
        else:
            for doc in X:
                yield [pos_tag(word_tokenize(sent)) for sent in sent_tokenize(doc)]


##########################################################################
## PosTag Tests
##########################################################################


class TestPosTag(VisualTestCase):
    """
    PosTag (Part of Speech Tagging Visualizer) Tests
    """

    def test_quick_method(self):
        """
        Assert no errors occur when using the quick method
        """
        # Fail if data hasn't been downloaded
        check_nltk_data()

        _, ax = plt.subplots()
        tagged_docs = list(get_tagged_docs(sonnets))

        viz = postag(tagged_docs, ax=ax, show=False)
        viz.ax.grid(False)

        # Fails on Miniconda/Appveyor with images not close (RMS 5.157)
        tol = 5.5 if IS_WINDOWS_OR_CONDA else 0.25
        self.assert_images_similar(viz, tol=tol)

    def test_unknown_tagset(self):
        """
        Ensure an exception is raised if the specified tagset is unknown
        """
        with pytest.raises(YellowbrickValueError):
            PosTagVisualizer(tagset="brill")

    def test_frequency_mode(self):
        """
        Assert no errors occur when the visualizer is run on frequency mode
        """
        check_nltk_data()

        _, ax = plt.subplots()
        tagged_docs = list(get_tagged_docs(sonnets))

        viz = PosTagVisualizer(ax=ax, frequency=True)
        viz.fit(tagged_docs)
        viz.finalize()
        ax.grid(False)

        # Sorted tags i.e predetermined order
        sorted_tags = [
            "noun",
            "adjective",
            "punctuation",
            "verb",
            "preposition",
            "determiner",
            "adverb",
            "conjunction",
            "pronoun",
            "wh- word",
            "modal",
            "infinitive",
            "possessive",
            "other",
            "symbol",
            "existential",
            "digit",
            "non-English",
            "interjection",
            "list",
        ]

        # Extract tick labels from the plot
        ticks_ax = [tick.get_text() for tick in ax.xaxis.get_ticklabels()]

        # Assert ticks are set properly (order may vary for equal frequencies)
        assert set(ticks_ax) == set(sorted_tags)

        # Fails on Miniconda/Appveyor with images not close (RMS 5.302)
        tol = 5.5 if IS_WINDOWS_OR_CONDA else 0.5
        self.assert_images_similar(ax=ax, tol=tol)

    @pytest.mark.skipif(nltk is None, reason="test requires nltk")
    def test_word_tagged(self):
        """
        Assert no errors occur during PosTagVisualizer integration
        with word tokenized corpus
        """
        # Fail if data hasn't been downloaded
        check_nltk_data()

        tagged_docs = list(get_tagged_docs(sonnets, model="nltk", tagger="word"))

        visualizer = PosTagVisualizer(tagset="penn_treebank")

        visualizer.fit(tagged_docs)
        visualizer.ax.grid(False)

        self.assert_images_similar(visualizer)

    @pytest.mark.skipif(nltk is None, reason="test requires nltk")
    def test_wordpunct_tagged(self):
        """
        Assert no errors occur during PosTagVisualizer integration
        with wordpunct tokenized corpus
        """
        # Fail if data hasn't been downloaded
        check_nltk_data()

        wordpunct_tagged_docs = list(
            get_tagged_docs(sonnets, model="nltk", tagger="wordpunct")
        )

        visualizer = PosTagVisualizer(tagset="penn_treebank")

        visualizer.fit(wordpunct_tagged_docs)
        visualizer.ax.grid(False)

        self.assert_images_similar(visualizer)

    @pytest.mark.skipif(spacy is None, reason="test requires spacy")
    def test_spacy_tagged(self):
        """
        Assert no errors occur during PosTagVisualizer integration
        with spacy tokenized corpus
        """
        # Fail if data hasn't been downloaded
        check_spacy_data()

        spacy_tagged_docs = list(get_tagged_docs(sonnets, model="spacy"))

        visualizer = PosTagVisualizer(tagset="universal")

        visualizer.fit(spacy_tagged_docs)
        visualizer.ax.grid(False)

        self.assert_images_similar(visualizer)

    @pytest.mark.skipif(spacy is None, reason="test requires spacy")
    def test_spacy_raw(self):
        """
        Assert no errors occur during PosTagVisualizer integration
        with raw corpus to be parsed using spacy
        """
        visualizer = PosTagVisualizer(parser="spacy", tagset="universal")
        visualizer.fit(sonnets)
        visualizer.ax.grid(False)

        self.assert_images_similar(visualizer)

    @pytest.mark.skipif(nltk is None, reason="test requires nltk")
    def test_nltk_word_raw(self):
        """
        Assert no errors occur during PosTagVisualizer integration
        with raw corpus to be parsed using nltk
        """
        visualizer = PosTagVisualizer(parser="nltk", tagset="penn_treebank")
        visualizer.fit(sonnets)
        visualizer.ax.grid(False)

        self.assert_images_similar(visualizer)

    @pytest.mark.skipif(nltk is None, reason="test requires nltk")
    def test_nltk_wordpunct_raw(self):
        """
        Assert no errors occur during PosTagVisualizer integration
        with raw corpus to be parsed using nltk
        """
        visualizer = PosTagVisualizer(parser="nltk_wordpunct", tagset="penn_treebank")
        visualizer.fit(sonnets)
        visualizer.ax.grid(False)

        self.assert_images_similar(visualizer)

    def test_stack_mode(self):
        """
        Assert no errors occur when the visualizer is run on stack mode
        """
        check_nltk_data()

        _, ax = plt.subplots()
        tagged_docs = list(get_tagged_docs(sonnets))

        visualizer = PosTagVisualizer(stack=True, ax=ax)
        visualizer.fit(tagged_docs, y=["a", "b", "c"])
        visualizer.ax.grid(False)

        self.assert_images_similar(ax=ax)

    def test_stack_frequency_mode(self):
        """
        Assert no errors occur when the visualizer is run on both stack and
        frequency mode
        """
        check_nltk_data()

        _, ax = plt.subplots()
        tagged_docs = list(get_tagged_docs(sonnets))

        visualizer = PosTagVisualizer(stack=True, frequency=True, ax=ax)
        visualizer.fit(tagged_docs, y=["a", "b", "c"])
        visualizer.ax.grid(False)

        # Sorted tags i.e predetermined order
        sorted_tags = [
            "noun",
            "adjective",
            "punctuation",
            "verb",
            "preposition",
            "determiner",
            "adverb",
            "conjunction",
            "pronoun",
            "wh- word",
            "modal",
            "infinitive",
            "possessive",
            "other",
            "symbol",
            "existential",
            "digit",
            "non-English",
            "interjection",
            "list",
        ]

        # Extract tick labels from the plot
        ticks_ax = [tick.get_text() for tick in ax.xaxis.get_ticklabels()]

        # Assert ticks are set properly (order may vary for equal frequencies)
        assert set(ticks_ax) == set(sorted_tags)

        # Image comparison may differ due to POS tag ordering variations
        self.assert_images_similar(ax=ax, tol=5.5)



#!/usr/bin/env python3
# validation_curve.py

"""
Generates the validation curve visualizations for the documentation
"""

##########################################################################
## Imports
##########################################################################

import numpy as np
import matplotlib.pyplot as plt

from sklearn.svm import SVC
from sklearn.datasets import load_digits
from sklearn.preprocessing import OneHotEncoder
from sklearn.neighbors import KNeighborsClassifier
from sklearn.model_selection import StratifiedKFold

from yellowbrick.datasets import load_game
from yellowbrick.model_selection import ValidationCurve


##########################################################################
## Helper Methods
##########################################################################


def validation_curve_sklearn_example(
    path="images/validation_curve_sklearn_example.png"
):
    digits = load_digits()
    X, y = digits.data, digits.target

    _, ax = plt.subplots()

    param_range = np.logspace(-6, -1, 5)
    oz = ValidationCurve(
        SVC(),
        ax=ax,
        param_name="gamma",
        param_range=param_range,
        logx=True,
        cv=10,
        scoring="accuracy",
        n_jobs=4,
    )
    oz.fit(X, y)
    oz.show(outpath=path)


def validation_curve_classifier_svc(path="images/validation_curve_classifier_svc.png"):
    X, y = load_game()
    X = OneHotEncoder().fit_transform(X)

    _, ax = plt.subplots()
    cv = StratifiedKFold(12)
    param_range = np.logspace(-6, -1, 12)

    print("warning: generating the SVC validation curve can take a very long time!")

    oz = ValidationCurve(
        SVC(),
        ax=ax,
        param_name="gamma",
        param_range=param_range,
        logx=True,
        cv=cv,
        scoring="f1_weighted",
        n_jobs=8,
    )
    oz.fit(X, y)
    oz.show(outpath=path)


def validation_curve_classifier_knn(path="images/validation_curve_classifier_knn.png"):
    X, y = load_game()
    X = OneHotEncoder().fit_transform(X)

    _, ax = plt.subplots()
    cv = StratifiedKFold(4)
    param_range = np.arange(3, 20, 2)

    print("warning: generating the KNN validation curve can take a very long time!")

    oz = ValidationCurve(
        KNeighborsClassifier(),
        ax=ax,
        param_name="n_neighbors",
        param_range=param_range,
        cv=cv,
        scoring="f1_weighted",
        n_jobs=8,
    )
    oz.fit(X, y)
    oz.show(outpath=path)


##########################################################################
## Main Method
##########################################################################

if __name__ == "__main__":
    # validation_curve_sklearn_example()
    validation_curve_classifier_svc()
    validation_curve_classifier_knn()



"""Test the RadViz feature analysis visualizers."""

import numpy as np
import pytest

from yellowbrick.exceptions import DataWarning
from yellowbrick.utils.nan_warnings import (
    count_nan_elements,
    count_rows_with_nans,
    warn_if_nans_exist,
    filter_missing,
)


def test_raise_warning_if_nans_exist():
    """Test that a warning is raised if any nans are in the data."""
    data = np.array([[1, 2, 3], [1, 2, np.nan]])

    with pytest.warns(DataWarning):
        warn_if_nans_exist(data)


def test_count_rows_in_2d_arrays_with_nans():
    """Test that nan-containinr rows in 2d arrays are counted correctly."""
    data_1_row = np.array([[1, 2, 3]])

    data_2_rows = np.array([[1, 2, 3], [1, 2, 3], [np.nan, 2, 3], [1, np.nan, 3]])

    data_3_rows = np.array(
        [[1, 2, 3], [np.nan, 2, 3], [1, np.nan, 3], [np.nan, np.nan, np.nan]]
    )

    assert count_rows_with_nans(data_1_row) == 0
    assert count_rows_with_nans(data_2_rows) == 2
    assert count_rows_with_nans(data_3_rows) == 3


def test_count_nan_elements():
    """Test that nan elements in 1d arrays are counted correctly."""
    data0 = np.array([1, 2, 3])
    data1 = np.array([1, np.nan, 3])
    data3 = np.array([np.nan, np.nan, np.nan])

    assert count_nan_elements(data0) == 0
    assert count_nan_elements(data1) == 1
    assert count_nan_elements(data3) == 3


def test_clean_data_X_only_no_nans():
    """Test that an array with no nulls is returned intact."""
    X = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])

    observed = filter_missing(X)
    np.testing.assert_array_equal(X, observed)


def test_clean_data_X_only():
    """Test that nan-containing X rows are removed without y."""
    X = np.array([[1, 2, np.nan], [4, 5, 6], [np.nan, np.nan, np.nan]])

    expected = np.array([[4, 5, 6]])
    observed = filter_missing(X)

    np.testing.assert_array_equal(expected, observed)


def test_clean_data_dirty_X_dirty_y():
    """Test that nan-containing X, y rows are removed when both contain nans."""
    X = np.array([[1, 2, 3], [4, 5, 6], [7, 8, np.nan], [np.nan, np.nan, np.nan]])
    y = np.array([33, np.nan, 44, np.nan])

    expected_X = np.array([[1, 2, 3]])
    expected_y = np.array([33])
    observed_X, observed_y = filter_missing(X, y)

    np.testing.assert_array_equal(expected_X, observed_X)
    np.testing.assert_array_equal(expected_y, observed_y)


def test_clean_data_dirty_X_clean_y():
    """Test that nan-containing X, y rows are removed when X contains nans."""
    X = np.array([[1, 2, 3], [4, 5, 6], [7, 8, np.nan], [np.nan, np.nan, np.nan]])
    y = np.array([33, 44, 55, 66])

    expected_X = np.array([[1, 2, 3], [4, 5, 6]])
    expected_y = np.array([33, 44])
    observed_X, observed_y = filter_missing(X, y)

    np.testing.assert_array_equal(expected_X, observed_X)
    np.testing.assert_array_equal(expected_y, observed_y)


def test_clean_data_clean_X_dirty_y():
    """Test that nan-containing X, y rows are removed when y contains nans."""
    X = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9], [10, 11, 12]])
    y = np.array([np.nan, 44, np.nan, 66])

    expected_X = np.array([[4, 5, 6], [10, 11, 12]])
    expected_y = np.array([44, 66])
    observed_X, observed_y = filter_missing(X, y)

    np.testing.assert_array_equal(expected_X, observed_X)
    np.testing.assert_array_equal(expected_y, observed_y)



# tests.fixtures
# Helpers for pytest fixtures and data related testing.
#
# Author:   Zijie (ZJ) Poh <zjpoh@noreply.github.com>
# Created:  Wed Feb 13 13:24:24 2019 -0400
#
# Copyright (C) 2019 The scikit-yb developers.
# For license information, see LICENSE.txt
#
# ID: fixtures.py [eb9f8cc] 8103276+zjpoh@users.noreply.github.com $

"""
Helpers for pytest fixtures and data related testing.
"""

##########################################################################
## Imports and Module Variables
##########################################################################

from collections import namedtuple


## Used for wrapping an dataset into a single variable.
Dataset = namedtuple("Dataset", "X,y")
Split = namedtuple("Split", "train,test")



#!/usr/bin/env python
# coding: utf-8

# # Random feature dropping curve
# 
# This notebook demonstrates the random feature dropping curve (also called a neuron dropping curve in neural decoding research).

# In[1]:


import matplotlib.pyplot as plt
import numpy as np

# Import scikit-learn utilities
from sklearn.pipeline import make_pipeline
from sklearn.ensemble import RandomForestClassifier
from sklearn.linear_model import LogisticRegression
from sklearn.naive_bayes import MultinomialNB
from sklearn.preprocessing import OneHotEncoder, LabelEncoder

# Import all of the Yellowbrick classifiers
from yellowbrick.datasets import load_game
from yellowbrick.model_selection.dropping_curve import  dropping_curve


def main():
    # Load Connect-4 game data
    X, y = load_game()

    print(f'X.shape={X.shape}')
    print(f'y.shape={y.shape}')

    X_enc = OneHotEncoder().fit_transform(X)
    le = LabelEncoder()
    y_enc = le.fit_transform(y)

    fig, ax = plt.subplots()
    dropping_curve(
        MultinomialNB(),
        X_enc,
        y_enc,
        feature_sizes=np.linspace(0.05, 1, 20),
        ax=ax,
    )


if __name__ == '__main__':
    main()



# tests.test_contrib.test_missing.test_bar
# Tests for the alpha selection visualizations.
#
# Author:  Nathan Danielsen <nathan.danielsen@gmail.com>
# Created:  Thu Mar 29 12:13:04 2018 -0500
#
# Copyright (C) 2018 The scikit-yb developers
# For license information, see LICENSE.txt
#
# ID: test_bar.py [1443e16] ndanielsen@users.noreply.github.com $

"""
Tests for the MissingValuesBar visualizations.
"""

##########################################################################
## Imports
##########################################################################

import os
import pytest
import numpy as np

from tests.base import VisualTestCase
from sklearn.datasets import make_classification
from yellowbrick.contrib.missing.bar import *

try:
    import pandas as pd
except ImportError:
    pd = None


@pytest.fixture(scope="class")
def missing_bar_tolerance(request):
    request.cls.tol = 0.5 if os.name == "nt" else 0.01


##########################################################################
## Feature Importances Tests
##########################################################################


@pytest.mark.usefixtures("missing_bar_tolerance")
class TestMissingBarVisualizer(VisualTestCase):
    """
    FeatureImportances visualizer
    """

    @pytest.mark.skipif(pd is None, reason="pandas is required")
    def test_missingvaluesbar_pandas(self):
        """
        Integration test of visualizer with pandas
        """
        X, y = make_classification(
            n_samples=400,
            n_features=20,
            n_informative=8,
            n_redundant=8,
            n_classes=2,
            n_clusters_per_class=4,
            random_state=854,
        )

        # add nan values to a range of values in the matrix
        X[X > 1.5] = np.nan
        X_ = pd.DataFrame(X)

        features = [str(n) for n in range(20)]
        viz = MissingValuesBar(features=features)
        viz.fit(X_)
        viz.finalize()
        
        self.assert_images_similar(viz, tol=self.tol)

    @pytest.mark.skipif(pd is None, reason="pandas is required")
    def test_missingvaluesbar_pandas_no_features_passed(self):
        """
        Integration test of visualizer with pandas
        """
        X, y = make_classification(
            n_samples=400,
            n_features=20,
            n_informative=8,
            n_redundant=8,
            n_classes=2,
            n_clusters_per_class=4,
            random_state=854,
        )

        # add nan values to a range of values in the matrix
        X[X > 1.5] = np.nan
        X_ = pd.DataFrame(X)

        viz = MissingValuesBar()
        viz.fit(X_)
        viz.finalize()

        self.assert_images_similar(viz, tol=self.tol)

    def test_missingvaluesbar_numpy(self):
        """
        Integration test of visualizer with numpy without target y passed in
        """
        X, y = make_classification(
            n_samples=400,
            n_features=20,
            n_informative=8,
            n_redundant=8,
            n_classes=2,
            n_clusters_per_class=4,
            random_state=856,
        )

        # add nan values to a range of values in the matrix
        X[X > 1.5] = np.nan

        features = [str(n) for n in range(20)]
        viz = MissingValuesBar(features=features)
        viz.fit(X)
        viz.finalize()

        self.assert_images_similar(viz, tol=self.tol)

    def test_missingvaluesbar_numpy_no_features_passed(self):
        """
        Integration test of visualizer with numpy without target y passed in
        """
        X, y = make_classification(
            n_samples=400,
            n_features=20,
            n_informative=8,
            n_redundant=8,
            n_classes=2,
            n_clusters_per_class=4,
            random_state=856,
        )

        # add nan values to a range of values in the matrix
        X[X > 1.5] = np.nan

        viz = MissingValuesBar()
        viz.fit(X)
        viz.finalize()

        self.assert_images_similar(viz, tol=self.tol)

    def test_missingvaluesbar_numpy_with_y_target(self):
        """
        Integration test of visualizer with numpy without target y passed in
        but no class labels
        """
        X, y = make_classification(
            n_samples=400,
            n_features=20,
            n_informative=8,
            n_redundant=8,
            n_classes=2,
            n_clusters_per_class=4,
            random_state=856,
        )

        # add nan values to a range of values in the matrix
        X[X > 1.5] = np.nan

        features = [str(n) for n in range(20)]
        viz = MissingValuesBar(features=features)
        viz.fit(X, y)
        viz.finalize()

        self.assert_images_similar(viz, tol=self.tol)

    def test_missingvaluesbar_numpy_with_y_target_with_labels(self):
        """
        Integration test of visualizer with numpy without target y passed in
        but no class labels
        """
        X, y = make_classification(
            n_samples=400,
            n_features=20,
            n_informative=8,
            n_redundant=8,
            n_classes=2,
            n_clusters_per_class=4,
            random_state=856,
        )

        # add nan values to a range of values in the matrix
        X[X > 1.5] = np.nan

        features = [str(n) for n in range(20)]
        viz = MissingValuesBar(features=features, classes=["class A", "class B"])
        viz.fit(X, y)
        viz.finalize()

        self.assert_images_similar(viz, tol=self.tol)

    def test_missingvaluesbar_numpy_with_string_and_bool_cols(self):
        """
        Integration test of visualizer with numpy array with string and boolean columns
        """
        X, y = make_classification(
            n_samples=400,
            n_features=10,
            n_informative=2,
            n_redundant=3,
            n_classes=2,
            n_clusters_per_class=2,
            random_state=854
        )

        # add nan values to a range of values in the matrix
        X[X > 1.5] = np.nan

        rng = np.random.default_rng(2021)
        fruit_choices = np.array(['apples', 'pears', 'peaches', "", np.nan, 'bananas'])
        fruits = rng.choice(fruit_choices, (400, 1))

        bool_choices = np.array([np.nan, False, True])
        booleans = rng.choice(bool_choices, (400, 1))

        X = np.append(X, fruits, axis=1)
        X = np.append(X, booleans, axis=1)

        features = [str(n) for n in range(12)]
        viz = MissingValuesBar(features=features)
        viz.fit(X, y)
        viz.finalize()

        self.assert_images_similar(viz, tol=5)

    @pytest.mark.skipif(pd is None, reason="pandas is required")
    def test_missingvaluesbar_pandas_with_string_and_bool_cols(self):
        """
        Integration test of visualizer with pandas dataframe with string and boolean columns
        """
        X, y = make_classification(
            n_samples=400,
            n_features=10,
            n_informative=2,
            n_redundant=3,
            n_classes=2,
            n_clusters_per_class=2,
            random_state=854
        )

        # add nan values to a range of values in the matrix
        X[X > 1.5] = np.nan

        rng = np.random.default_rng(2021)
        fruit_choices = np.array(['apples', 'pears', 'peaches', "", np.nan, 'bananas'])
        fruits = rng.choice(fruit_choices, (400, 1))

        bool_choices = np.array([np.nan, False, True])
        booleans = rng.choice(bool_choices, (400, 1))

        X = np.append(X, fruits, axis=1)
        X = np.append(X, booleans, axis=1)

        X_ = pd.DataFrame(X)

        features = [str(n) for n in range(12)]
        viz = MissingValuesBar(features=features)
        viz.fit(X_, y)
        viz.finalize()

        self.assert_images_similar(viz, tol=5)

    def test_missingvaluesbar_numpy_with_mixed_dtypes(self):
        """
        Integration test of visualizer with numpy array with mixed dtypes in a single column
        """
        X, y = make_classification(
            n_samples=400,
            n_features=10,
            n_informative=2,
            n_redundant=3,
            n_classes=2,
            n_clusters_per_class=2,
            random_state=854
        )

        # add nan values to a range of values in the matrix
        X[X > 1.5] = np.nan

        rng = np.random.default_rng(2021)
        mixed_dtype_choices = np.array(['apples', 'pears', 'peaches', "", np.nan, 'bananas', 1, 2.4, 5.6, False, True])
        mixed_dtypes = rng.choice(mixed_dtype_choices, (400, 1))

        X_with_mixed_dtypes = np.append(X, mixed_dtypes, axis=1)

        features = [str(n) for n in range(11)]
        viz = MissingValuesBar(features=features)
        viz.fit(X_with_mixed_dtypes, y)
        viz.finalize()

        self.assert_images_similar(viz, tol=5)

    @pytest.mark.skipif(pd is None, reason="pandas is required")
    def test_missingvaluesbar_pandas_with_mixed_dtypes(self):
        """
        Integration test of visualizer with pandas dataframe with mixed dtypes in a single column
        """
        X, y = make_classification(
            n_samples=400,
            n_features=10,
            n_informative=2,
            n_redundant=3,
            n_classes=2,
            n_clusters_per_class=2,
            random_state=854
        )

        # add nan values to a range of values in the matrix
        X[X > 1.5] = np.nan

        rng = np.random.default_rng(2021)
        mixed_dtype_choices = np.array(['apples', 'pears', 'peaches', "", np.nan, 'bananas', 1, 2.4, 5.6, False, True])
        mixed_dtypes = rng.choice(mixed_dtype_choices, (400, 1))

        X_with_mixed_dtypes = np.append(X, mixed_dtypes, axis=1)
        X_ = pd.DataFrame(X_with_mixed_dtypes)

        features = [str(n) for n in range(11)]
        viz = MissingValuesBar(features=features)
        viz.fit(X_, y)
        viz.finalize()

        self.assert_images_similar(viz, tol=5)



# tests.test_model_selection.conftest
# Provides fixtures for the model selection tests module.
#
# Author:  Benjamin Bengfort <benjamin@bengfort.com>
# Created: Fri Mar 30 14:35:39 2018 -0400
#
# ID: conftest.py [c5355ee] benjamin@bengfort.com $

"""
Provides fixtures for the classification tests module.
"""

##########################################################################
## Imports
##########################################################################

import pytest

from tests.fixtures import Dataset
from sklearn.datasets import make_classification, make_regression, make_blobs


##########################################################################
## Fixtures
##########################################################################


@pytest.fixture(scope="class")
def classification(request):
    """
    Creates a random multiclass classification dataset fixture
    """
    X, y = make_classification(
        n_samples=500,
        n_features=20,
        n_informative=8,
        n_redundant=2,
        n_classes=3,
        n_clusters_per_class=3,
        random_state=3902,
    )

    dataset = Dataset(X, y)
    request.cls.classification = dataset


@pytest.fixture(scope="class")
def regression(request):
    """
    Creates a random regression dataset fixture
    """
    X, y = make_regression(
        n_samples=500,
        n_features=20,
        n_informative=8,
        noise=0.01,
        bias=1.4,
        random_state=953,
    )

    dataset = Dataset(X, y)
    request.cls.regression = dataset


@pytest.fixture(scope="class")
def clusters(request):
    """
    Creates a random regression dataset fixture
    """
    X, y = make_blobs(n_samples=500, n_features=20, centers=3, random_state=743)

    dataset = Dataset(X, y)
    request.cls.clusters = dataset



# tests.test_features.test_pcoords
# Testing for the parallel coordinates feature visualizers
#
# Author:  Benjamin Bengfort
# Author:  @thekylesaurus
# Created: Thu Oct 06 11:21:27 2016 -0400
#
# Copyright (C) 2017 The scikit-yb developers.
# For license information, see LICENSE.txt
#
# ID: test_pcoords.py [1d407ab] benjamin@bengfort.com $

"""
Testing for the parallel coordinates feature visualizers
"""

##########################################################################
## Imports
##########################################################################

import pytest
import numpy as np

from yellowbrick.datasets import load_occupancy
from yellowbrick.features.pcoords import *

from tests.base import VisualTestCase
from ..fixtures import Dataset
from sklearn.datasets import make_classification


try:
    import pandas as pd
except ImportError:
    pd = None


##########################################################################
## Fixtures
##########################################################################


@pytest.fixture(scope="class")
def dataset(request):
    """
    Creates a random multiclass classification dataset fixture
    """
    X, y = make_classification(
        n_samples=200,
        n_features=5,
        n_informative=4,
        n_redundant=0,
        n_classes=3,
        n_clusters_per_class=1,
        random_state=451,
        flip_y=0,
        class_sep=3,
        scale=np.array([1.0, 2.0, 100.0, 20.0, 1.0]),
    )

    dataset = Dataset(X, y)
    request.cls.dataset = dataset


##########################################################################
## Parallel Coordinates Tests
##########################################################################


@pytest.mark.usefixtures("dataset")
class TestParallelCoordinates(VisualTestCase):
    """
    Test the ParallelCoordinates visualizer
    """

    def test_parallel_coords(self):
        """
        Test images closeness on random 3 class dataset
        """
        visualizer = ParallelCoordinates()
        visualizer.fit_transform(self.dataset.X, self.dataset.y)
        visualizer.finalize()
        self.assert_images_similar(visualizer, tol=0.25)

    def test_parallel_coords_fast(self):
        """
        Test images closeness on random 3 class dataset in fast mode
        """
        visualizer = ParallelCoordinates(fast=True)
        visualizer.fit_transform(self.dataset.X, self.dataset.y)
        visualizer.finalize()
        self.assert_images_similar(visualizer, tol=0.25)

    def test_alpha(self):
        """
        Test image closeness on opaque alpha for random 3 class dataset
        """
        visualizer = ParallelCoordinates(alpha=1.0)
        visualizer.fit_transform(self.dataset.X, self.dataset.y)
        visualizer.finalize()
        self.assert_images_similar(visualizer, tol=0.25)

    def test_alpha_fast(self):
        """
        Test image closeness on opaque alpha for random 3 class dataset in fast mode
        """
        visualizer = ParallelCoordinates(alpha=1.0, fast=True)
        visualizer.fit_transform(self.dataset.X, self.dataset.y)
        visualizer.finalize()
        self.assert_images_similar(visualizer, tol=0.25)

    def test_labels(self):
        """
        Test image closeness when class and feature labels are supplied
        """
        visualizer = ParallelCoordinates(
            classes=["a", "b", "c"], features=["f1", "f2", "f3", "f4", "f5"]
        )
        visualizer.fit_transform(self.dataset.X, self.dataset.y)
        visualizer.finalize()
        self.assert_images_similar(visualizer)

    def test_labels_fast(self):
        """
        Test image closeness when class and feature labels are supplied in fast mode
        """
        visualizer = ParallelCoordinates(
            classes=["a", "b", "c"], features=["f1", "f2", "f3", "f4", "f5"], fast=True
        )
        visualizer.fit_transform(self.dataset.X, self.dataset.y)
        visualizer.finalize()
        self.assert_images_similar(visualizer)

    def test_normalized_l2(self):
        """
        Test image closeness on l2 normalized 3 class dataset
        """
        visualizer = ParallelCoordinates(normalize="l2")
        visualizer.fit_transform(self.dataset.X, self.dataset.y)
        visualizer.finalize()
        self.assert_images_similar(visualizer, tol=0.25)

    def test_normalized_l2_fast(self):
        """
        Test image closeness on l2 normalized 3 class dataset in fast mode
        """
        visualizer = ParallelCoordinates(normalize="l2", fast=True)
        visualizer.fit_transform(self.dataset.X, self.dataset.y)
        visualizer.finalize()
        self.assert_images_similar(visualizer, tol=0.25)

    def test_normalized_minmax(self):
        """
        Test image closeness on minmax normalized 3 class dataset
        """
        visualizer = ParallelCoordinates(normalize="minmax")
        visualizer.fit_transform(self.dataset.X, self.dataset.y)
        visualizer.finalize()
        self.assert_images_similar(visualizer, tol=0.25)

    def test_normalized_minmax_fast(self):
        """
        Test image closeness on minmax normalized 3 class dataset in fast mode
        """
        visualizer = ParallelCoordinates(normalize="minmax", fast=True)
        visualizer.fit_transform(self.dataset.X, self.dataset.y)
        visualizer.finalize()
        self.assert_images_similar(visualizer, tol=0.25)

    def test_parallel_coordinates_quickmethod(self):
        """
        Test the quick method producing a valid visualization
        """
        X, y = load_occupancy(return_dataset=True).to_numpy()

        # Compare the images
        # Use only the first 100 samples so the test will run faster
        visualizer = parallel_coordinates(X, y, sample=100, show=False)
        self.assert_images_similar(visualizer)

    @pytest.mark.skipif(pd is None, reason="test requires pandas")
    def test_pandas_integration_sampled(self):
        """
        Test on a real dataset with pandas DataFrame and Series sampled for speed
        """
        data = load_occupancy(return_dataset=True)
        X, y = data.to_pandas()
        classes = [
            k for k, _ in sorted(data.meta["labels"].items(), key=lambda i: i[1])
        ]

        assert isinstance(X, pd.DataFrame)
        assert isinstance(y, pd.Series)

        oz = ParallelCoordinates(
            sample=0.05, shuffle=True, random_state=4291, classes=classes
        )
        oz.fit_transform(X, y)
        oz.finalize()

        self.assert_images_similar(oz, tol=0.1)

    def test_numpy_integration_sampled(self):
        """
        Ensure visualizer works in default case with numpy arrays and sampling
        """
        data = load_occupancy(return_dataset=True)
        X, y = data.to_numpy()
        classes = [
            k for k, _ in sorted(data.meta["labels"].items(), key=lambda i: i[1])
        ]

        assert isinstance(X, np.ndarray)
        assert isinstance(y, np.ndarray)

        oz = ParallelCoordinates(
            sample=0.05, shuffle=True, random_state=4291, classes=classes
        )
        oz.fit_transform(X, y)
        oz.finalize()

        self.assert_images_similar(oz, tol=0.1)

    @pytest.mark.skipif(pd is None, reason="test requires pandas")
    def test_pandas_integration_fast(self):
        """
        Test on a real dataset with pandas DataFrame and Series in fast mode
        """
        data = load_occupancy(return_dataset=True)
        X, y = data.to_pandas()
        classes = [
            k for k, _ in sorted(data.meta["labels"].items(), key=lambda i: i[1])
        ]

        assert isinstance(X, pd.DataFrame)
        assert isinstance(y, pd.Series)

        oz = ParallelCoordinates(fast=True, classes=classes)
        oz.fit_transform(X, y)
        oz.finalize()

        self.assert_images_similar(oz, tol=0.1)

    def test_numpy_integration_fast(self):
        """
        Ensure visualizer works in default case with numpy arrays and fast mode
        """
        data = load_occupancy(return_dataset=True)
        X, y = data.to_numpy()
        classes = [
            k for k, _ in sorted(data.meta["labels"].items(), key=lambda i: i[1])
        ]

        assert isinstance(X, np.ndarray)
        assert isinstance(y, np.ndarray)

        oz = ParallelCoordinates(fast=True, classes=classes)
        oz.fit_transform(X, y)
        oz.finalize()

        self.assert_images_similar(oz, tol=0.1)

    def test_normalized_invalid_arg(self):
        """
        Invalid argument to 'normalize' should raise
        """
        with pytest.raises(YellowbrickValueError):
            ParallelCoordinates(normalize="foo")

    def test_sample_int(self):
        """
        Assert no errors occur using integer 'sample' argument
        """
        visualizer = ParallelCoordinates(sample=10)
        visualizer.fit_transform(self.dataset.X, self.dataset.y)

    def test_sample_int_shuffle(self):
        """
        Assert no errors occur using integer 'sample' argument and shuffle, with different random_state args
        """
        visualizer = ParallelCoordinates(sample=3, shuffle=True)
        visualizer.fit_transform(self.dataset.X, self.dataset.y)

        visualizer = ParallelCoordinates(sample=3, shuffle=True, random_state=444)
        visualizer.fit_transform(self.dataset.X, self.dataset.y)

        visualizer = ParallelCoordinates(
            sample=3, shuffle=True, random_state=np.random.RandomState()
        )
        visualizer.fit_transform(self.dataset.X, self.dataset.y)

    def test_sample_int_shuffle_false(self):
        """
        Assert no errors occur using integer 'sample' argument and shuffle, with different random_state args
        """
        visualizer = ParallelCoordinates(sample=3, shuffle=False)
        visualizer.fit_transform(self.dataset.X, self.dataset.y)

        visualizer = ParallelCoordinates(sample=3, shuffle=False, random_state=444)
        visualizer.fit_transform(self.dataset.X, self.dataset.y)

        visualizer = ParallelCoordinates(
            sample=3, shuffle=False, random_state=np.random.RandomState()
        )
        visualizer.fit_transform(self.dataset.X, self.dataset.y)

    def test_sample_int_invalid(self):
        """
        Negative int values should raise exception
        """
        with pytest.raises(YellowbrickValueError):
            ParallelCoordinates(sample=-1)

    def test_sample_float(self):
        """
        Assert no errors occur using float 'sample' argument
        """
        visualizer = ParallelCoordinates(sample=0.5)
        visualizer.fit_transform(self.dataset.X, self.dataset.y)

    def test_sample_float_shuffle(self):
        """
        Assert no errors occur using float 'sample' argument and shuffle, with different random_state args
        """
        visualizer = ParallelCoordinates(sample=0.5, shuffle=True)
        visualizer.fit_transform(self.dataset.X, self.dataset.y)

        visualizer = ParallelCoordinates(sample=0.5, shuffle=True, random_state=444)
        visualizer.fit_transform(self.dataset.X, self.dataset.y)

        visualizer = ParallelCoordinates(
            sample=0.5, shuffle=True, random_state=np.random.RandomState()
        )
        visualizer.fit_transform(self.dataset.X, self.dataset.y)

    def test_sample_float_shuffle_false(self):
        """
        Assert no errors occur using float 'sample' argument and shuffle, with different random_state args
        """
        visualizer = ParallelCoordinates(sample=0.5, shuffle=False)
        visualizer.fit_transform(self.dataset.X, self.dataset.y)

        visualizer = ParallelCoordinates(sample=0.5, shuffle=False, random_state=444)
        visualizer.fit_transform(self.dataset.X, self.dataset.y)

        visualizer = ParallelCoordinates(
            sample=0.5, shuffle=False, random_state=np.random.RandomState()
        )
        visualizer.fit_transform(self.dataset.X, self.dataset.y)

    def test_sample_float_invalid(self):
        """
        Float values for 'sample' argument outside [0,1] should raise.
        """
        with pytest.raises(YellowbrickValueError):
            ParallelCoordinates(sample=-0.2)

        with pytest.raises(YellowbrickValueError):
            ParallelCoordinates(sample=1.1)

    def test_sample_invalid_type(self):
        """
        Non-numeric values for 'sample' argument should raise.
        """
        with pytest.raises(YellowbrickTypeError):
            ParallelCoordinates(sample="foo")

    @staticmethod
    def test_static_subsample():
        """
        Assert output of subsampling method against expectations
        """

        ntotal = 100
        ncols = 50

        y = np.arange(ntotal)
        X = np.ones((ntotal, ncols)) * y.reshape(ntotal, 1)

        visualizer = ParallelCoordinates(sample=1.0, random_state=None, shuffle=False)
        Xprime, yprime = visualizer._subsample(X, y)
        assert np.array_equal(Xprime, X)
        assert np.array_equal(yprime, y)

        visualizer = ParallelCoordinates(sample=200, random_state=None, shuffle=False)
        Xprime, yprime = visualizer._subsample(X, y)
        assert np.array_equal(Xprime, X)
        assert np.array_equal(yprime, y)

        sample = 50
        visualizer = ParallelCoordinates(
            sample=sample, random_state=None, shuffle=False
        )
        Xprime, yprime = visualizer._subsample(X, y)
        assert np.array_equal(Xprime, X[:sample, :])
        assert np.array_equal(yprime, y[:sample])

        sample = 50
        visualizer = ParallelCoordinates(sample=sample, random_state=None, shuffle=True)
        Xprime, yprime = visualizer._subsample(X, y)
        assert np.array_equal(Xprime, X[yprime.flatten(), :])
        assert len(Xprime) == sample
        assert len(yprime) == sample

        visualizer = ParallelCoordinates(sample=0.5, random_state=None, shuffle=False)
        Xprime, yprime = visualizer._subsample(X, y)
        assert np.array_equal(Xprime, X[: int(ntotal / 2), :])
        assert np.array_equal(yprime, y[: int(ntotal / 2)])

        sample = 0.5
        visualizer = ParallelCoordinates(sample=sample, random_state=None, shuffle=True)
        Xprime, yprime = visualizer._subsample(X, y)
        assert np.array_equal(Xprime, X[yprime.flatten(), :])
        assert len(Xprime) == ntotal * sample
        assert len(yprime) == ntotal * sample

        sample = 0.25
        visualizer = ParallelCoordinates(sample=sample, random_state=444, shuffle=True)
        Xprime, yprime = visualizer._subsample(X, y)
        assert np.array_equal(Xprime, X[yprime.flatten(), :])
        assert len(Xprime) == ntotal * sample
        assert len(yprime) == ntotal * sample

        sample = 0.99
        visualizer = ParallelCoordinates(
            sample=sample, random_state=np.random.RandomState(), shuffle=True
        )
        Xprime, yprime = visualizer._subsample(X, y)
        assert np.array_equal(Xprime, X[yprime.flatten(), :])
        assert len(Xprime) == ntotal * sample
        assert len(yprime) == ntotal * sample



##########################################################################
## Imports
##########################################################################

import pytest

from yellowbrick.exceptions import NotFitted


##########################################################################
## NotFitted Exception Tests
##########################################################################


class TestExceptions(object):
    """
    Test exception specific code and utilities
    """

    @pytest.mark.parametrize("method", ["transform", None])
    def test_not_fitted_from_estimator(self, method):
        """
        Ensure not fitted can be raised directly from an estimator
        """
        msg = "instance is not fitted yet, please call fit"
        with pytest.raises(NotFitted, match=msg):
            raise NotFitted.from_estimator(self, method)



# tests.test_utils.test_timer
# Tests for the stand alone timer functions in Yellowbrick utils.
#
# Author:   ZJ Poh <poh.zijie@gmail.com>
# Created:  Tue Jul 17 21:11:11 2018 -0700
#
# Copyright (C) 2017 The scikit-yb developers
# For license information, see LICENSE.txt
"""
Tests for the stand alone timer functions in Yellowbrick utils.
"""

##########################################################################
## Imports
##########################################################################

import pytest

from unittest import mock
from yellowbrick.utils.timer import *


##########################################################################
## Helper Function Tests
##########################################################################


class TestTimer(object):
    """
    Timer functions and utilities
    """

    @mock.patch("time.time", mock.Mock(side_effect=[1234.2, 1242.8]))
    def test_timer(self):
        with Timer() as timer:
            pass
        assert isinstance(timer.interval, float)
        assert timer.interval == pytest.approx(8.6)


@pytest.mark.parametrize(
    "s,expected",
    [
        (1.01, "00:00:01.0100"),
        (61.01, "00:01:01.0100"),
        (3661.01, "01:01:01.0100"),
        (360061.01, "100:01:01.0100"),
    ],
)
def test_human_readable_time(s, expected):
    assert human_readable_time(s) == expected



# tests.conftest
# Global definitions for Yellowbrick pytest
#
# Author:  Benjamin Bengfort <bbengfort@districtdatalabs.com>
# Created: Fri Mar 02 11:53:55 2018 -0500
#
# Copyright (C) 2016 The scikit-yb developers
# For license information, see LICENSE.txt
#
# ID: conftest.py [957cd53] benjamin@bengfort.com $

"""
Global definitions for Yellowbrick PyTest
"""

##########################################################################
## Imports
##########################################################################

import os
import matplotlib as mpl

from pytest_flakes import FlakesItem


##########################################################################
## Configure tests
##########################################################################


def pytest_configure(config):
    """
    This function is called by pytest for every plugin and conftest file
    after the command line arguments have been passed but before the
    session object is created and all of the tests are created. It is used
    to set a global configuration before all tests are run.

    Yellowbrick uses this function primarily to ensure that the matplotlib
    environment is setup correctly for all tests.
    """
    # This is redundant with the line in tests/__init__.py but ensures that
    # the backend is correctly set across all tests and plugins.
    mpl.use("Agg")

    # Travis-CI does not have san-serif so ensure standard fonts are used.
    # TODO: this is currently being reset before each test; needs fixing.
    mpl.rcParams["font.family"] = "DejaVu Sans"

##########################################################################
## PyTest Hooks
##########################################################################


def docline(obj):
    """
    Returns the first line of the object's docstring or None if
    there is no __doc__ on the object.
    """
    if not obj.__doc__:
        return None
    lines = list(filter(None, obj.__doc__.split("\n")))
    return lines[0].strip()


def pytest_itemcollected(item):
    """
    A reporting hook that is called when a test item is collected.

    This function can do many things to modify test output and test
    handling and in the future can be modified with those tasks. Right now
    we're currently only using this function to create node ids that
    pytest-spec knows how to parse and display as spec-style output.

    .. seealso:: https://stackoverflow.com/questions/28898919/use-docstrings-to-list-tests-in-py-test
    """

    # Ignore Session and PyFlake tests that are generated automatically
    if not hasattr(item.parent, "obj") or isinstance(item, FlakesItem):
        return

    # Collect test objects to inspect
    parent = item.parent.obj
    node = item.obj

    # Goal: produce a parsable string of the relative path, parent docstring
    # or class name, and the docstring of the test case, then set the nodeid
    # so that pytest-spec will correctly parse the information.
    path = os.path.relpath(str(item.fspath))
    prefix = docline(parent) or getattr(parent, "__name__", parent.__class__.__name__)
    suffix = docline(node) or node.__name__

    # Add parametrize or test generation id to distinguish it in output
    # TODO: this is broken with pytest 4.2 (no attribute _genid)
    if hasattr(item, "_genid") and item._genid:
        suffix += " ({})".format(item._genid)

    if prefix or suffix:
        item._nodeid = "::".join((path, prefix.strip(), suffix.strip()))



# tests.test_contrib.test_prepredict
# Test the prepredict estimator.
#
# Author:  Benjamin Bengfort <benjamin@bengfort.com>
# Created: Mon Jul 12 07:07:33 2021 -0400
#
# ID: test_prepredict.py [] benjamin@bengfort.com $

"""
Test the prepredict estimator.
"""

##########################################################################
## Imports
##########################################################################

import pytest

from io import BytesIO
from tests.fixtures import Dataset, Split
from tests.base import IS_WINDOWS_OR_CONDA, VisualTestCase

from sklearn.naive_bayes import GaussianNB
from sklearn.cluster import MiniBatchKMeans
from sklearn.linear_model import LinearRegression
from sklearn.model_selection import train_test_split as tts
from sklearn.datasets import make_classification, make_regression, make_blobs

from yellowbrick.contrib.prepredict import *
from yellowbrick.regressor import PredictionError
from yellowbrick.classifier import ClassificationReport
import numpy as np

# Set random state
np.random.seed()

##########################################################################
## Fixtures
##########################################################################


@pytest.fixture(scope="class")
def multiclass(request):
    """
    Creates a random multiclass classification dataset fixture
    """
    X, y = make_classification(
        n_samples=500,
        n_features=20,
        n_informative=8,
        n_redundant=2,
        n_classes=6,
        n_clusters_per_class=3,
        random_state=87,
    )

    X_train, X_test, y_train, y_test = tts(X, y, test_size=0.2, random_state=93)

    dataset = Dataset(Split(X_train, X_test), Split(y_train, y_test))
    request.cls.multiclass = dataset


@pytest.fixture(scope="class")
def continuous(request):
    """
    Creates a random continuous regression dataset fixture
    """
    X, y = make_regression(
        n_samples=500,
        n_features=22,
        n_informative=8,
        random_state=42,
        noise=0.2,
        bias=0.2,
    )

    X_train, X_test, y_train, y_test = tts(X, y, test_size=0.2, random_state=11)

    # Set a class attribute for regression
    request.cls.continuous = Dataset(Split(X_train, X_test), Split(y_train, y_test))


@pytest.fixture(scope="class")
def blobs(request):
    """
    Create a random blobs clustering dataset fixture
    """
    X, y = make_blobs(
        n_samples=1000, n_features=12, centers=6, shuffle=True, random_state=42
    )

    # Set a class attribute for blobs
    request.cls.blobs = Dataset(X, y)


##########################################################################
## Tests
##########################################################################


@pytest.mark.usefixtures("multiclass")
@pytest.mark.usefixtures("continuous")
@pytest.mark.usefixtures("blobs")
class TestPrePrePredictEstimator(VisualTestCase):
    """
    Pre-predict contrib tests.
    """

    @pytest.mark.xfail(
        IS_WINDOWS_OR_CONDA,
        reason="image comparison failure on Conda 3.8 and 3.9 with RMS 19.307",
    )
    def test_prepredict_classifier(self):
        """
        Test the prepredict estimator with classification report
        """
        # Make prepredictions
        X, y = self.multiclass.X, self.multiclass.y
        y_pred = GaussianNB().fit(X.train, y.train).predict(X.test)

        # Create prepredict estimator with prior predictions
        estimator = PrePredict(y_pred, CLASSIFIER)
        assert estimator.fit(X.train, y.train) is estimator
        assert estimator.predict(X.train) is y_pred
        assert estimator.score(X.test, y.test) == pytest.approx(0.41, rel=1e-3)

        # Test that a visualizer works with the pre-predictions.
        viz = ClassificationReport(estimator)
        viz.fit(None, y.train)
        viz.score(None, y.test)
        viz.finalize()

        self.assert_images_similar(viz)

    def test_prepredict_regressor(self):
        """
        Test the prepredict estimator with a prediction error plot
        """
        # Make prepredictions
        X, y = self.continuous.X, self.continuous.y
        y_pred = LinearRegression().fit(X.train, y.train).predict(X.test)

        # Create prepredict estimator with prior predictions
        estimator = PrePredict(y_pred, REGRESSOR)
        assert estimator.fit(X.train, y.train) is estimator
        assert estimator.predict(X.train) is y_pred
        assert estimator.score(X.test, y.test) == pytest.approx(
            0.9999983124154966, rel=1e-2
        )

        # Test that a visualizer works with the pre-predictions.
        viz = PredictionError(estimator)
        viz.fit(X.train, y.train)
        viz.score(X.test, y.test)
        viz.finalize()

        self.assert_images_similar(viz, tol=10.0)

    def test_prepredict_clusterer(self):
        """
        Test the prepredict estimator with a silhouette visualizer
        """
        X = self.blobs.X
        y_pred = MiniBatchKMeans(random_state=831).fit(X).predict(X)

        # Create prepredict estimator with prior predictions
        estimator = PrePredict(y_pred, CLUSTERER)
        assert estimator.fit(X) is estimator
        assert estimator.predict(X) is y_pred
        # Increase tolerance to 5% for scikit-learn algorithm changes
        assert estimator.score(X) == pytest.approx(0.5477478541994333, rel=5e-2)

        # NOTE: there is currently no cluster visualizer that can take advantage of
        # the prepredict utility since they all require learned attributes.

    def test_load(self):
        """
        Test the various ways that prepredict loads data
        """
        # Test callable
        ppe = PrePredict(lambda: self.multiclass.y.test)
        assert ppe._load() is self.multiclass.y.test

        # Test file-like object, assume that str and pathlib.Path work similarly
        f = BytesIO()
        np.save(f, self.continuous.y.test)
        f.seek(0)
        ppe = PrePredict(f)
        assert np.array_equal(ppe._load(), self.continuous.y.test)

        # Test direct array-like completed in other tests.



# tests.test_style
# Tests for the style handling module of yellowbrick.
#
# Author:   Benjamin Bengfort <bbengfort@districtdatalabs.com>
# Created:  Tue Oct 04 16:21:21 2016 -0400
#
# Copyright (C) 2016 The scikit-yb developers
# For license information, see LICENSE.txt
#
# ID: __init__.py [c6aff34] benjamin@bengfort.com $

"""
Tests for the style handling module of yellowbrick.
"""

##########################################################################
## Imports
##########################################################################



import time
import matplotlib.pyplot as plt
from sklearn.datasets import load_iris
from yellowbrick.features import ParallelCoordinates
import pandas as pd
import numpy as np


def plot_speedup(trials=5, factors=np.arange(1, 11)):
    def pcoords_time(X, y, fast=True):
        _, ax = plt.subplots()
        oz = ParallelCoordinates(fast=fast, ax=ax)

        start = time.time()
        oz.fit_transform(X, y)
        delta = time.time() - start

        plt.cla()  # clear current axis
        plt.clf()  # clear current figure
        plt.close("all")  # close all existing plots

        return delta

    def pcoords_speedup(X, y):
        fast_time = pcoords_time(X, y, fast=True)
        slow_time = pcoords_time(X, y, fast=False)

        return slow_time / fast_time

    data = load_iris()

    speedups = []
    variance = []

    for factor in factors:
        X = np.repeat(data.data, factor, axis=0)
        y = np.repeat(data.target, factor, axis=0)

        local_speedups = []
        for trial in range(trials):
            local_speedups.append(pcoords_speedup(X, y))

        local_speedups = np.array(local_speedups)
        speedups.append(local_speedups.mean())
        variance.append(local_speedups.std())

    speedups = np.array(speedups)
    variance = np.array(variance)

    series = pd.Series(speedups, index=factors)
    _, ax = plt.subplots(figsize=(9, 6))
    series.plot(ax=ax, marker="o", label="speedup factor", color="b")

    # Plot one standard deviation above and below the mean
    ax.fill_between(
        factors, speedups - variance, speedups + variance, alpha=0.25, color="b"
    )

    ax.set_ylabel("speedup factor")
    ax.set_xlabel("dataset size (number of repeats in Iris dataset)")
    ax.set_title("Speed Improvement of Fast Parallel Coordinates")
    plt.savefig("images/fast_parallel_coordinates_speedup_benchmark.png")


if __name__ == "__main__":
    plot_speedup()



# tests.test_text.test_freqdist
# Tests for the frequency distribution visualization
#
# Author:   Rebecca Bilbro
# Created:  2017-03-22 15:27
#
# Copyright (C) 2018 The scikit-yb developers
# For license information, see LICENSE.txt
#
# ID: test_freqdist.py [bd9cbb9] rebecca.bilbro@bytecubed.com $

"""
Tests for the frequency distribution text visualization
"""

##########################################################################
## Imports
##########################################################################

import pytest

from yellowbrick.datasets import load_hobbies
from yellowbrick.text.freqdist import *
from tests.base import IS_WINDOWS_OR_CONDA, VisualTestCase

from sklearn.feature_extraction.text import CountVectorizer

##########################################################################
## Data
##########################################################################

corpus = load_hobbies()

##########################################################################
## FreqDist Tests
##########################################################################


class TestFreqDist(VisualTestCase):
    @pytest.mark.xfail(
        IS_WINDOWS_OR_CONDA,
        reason="font rendering different in OS and/or Python; see #892",
    )
    def test_integrated_freqdist(self):
        """
        Assert no errors occur during freqdist integration
        """
        vectorizer = CountVectorizer()

        docs = vectorizer.fit_transform(corpus.data)
        features = vectorizer.get_feature_names_out()

        visualizer = FreqDistVisualizer(features)
        visualizer.fit(docs)

        visualizer.finalize()
        self.assert_images_similar(visualizer, tol=0.5)  # w/o tol fails with RMS 0.121

    def test_freqdist_quickmethod(self):
        """
        Assert no errors occur during freqdist quickmethod
        """
        vectorizer = CountVectorizer()

        docs = vectorizer.fit_transform(corpus.data)
        features = vectorizer.get_feature_names_out()

        viz = freqdist(features, docs, show=False)

        assert isinstance(viz, FreqDistVisualizer)

        # yellowbrick.exceptions.ImageComparisonFailure: images not close (RMS 1.401)
        self.assert_images_similar(viz, tol=1.5)



# tests.test_features.conftest
# Provides fixtures for the feature tests module.
#
# Author:   Naresh Bachwani
# Created:  Thu Aug 15 07:35:53 2019 -0400
#
# Copyright (C) 2016 The scikit-yb developers
# For license information, see LICENSE.txt
#
# ID: conftest.py [2c5f0e9] 43993586+naresh-bachwani@users.noreply.github.com $

"""
Provides fixtures for the feature tests module.
"""

##########################################################################
## Imports
##########################################################################

import pytest

from tests.fixtures import Dataset
from sklearn.datasets import make_classification, make_regression, make_s_curve


##########################################################################
## Fixtures
##########################################################################


@pytest.fixture(scope="class")
def discrete(request):
    """
    Creates a fixture of train and test splits for the sklearn digits dataset
    For ease of use returns a Dataset named tuple composed of two Split tuples.
    """
    X, y = make_classification(
        n_samples=400,
        n_features=12,
        n_informative=8,
        n_redundant=0,
        n_classes=5,
        n_clusters_per_class=1,
        class_sep=1.8,
        random_state=854,
        scale=[14.2, 2.1, 0.32, 0.001, 32.3, 44.1, 102.3, 2.3, 2.4, 38.2, 0.05, 1.0],
    )

    # Set a class attribute for discrete data.
    request.cls.discrete = Dataset(X, y)


@pytest.fixture(scope="class")
def continuous(request):
    """
    Creates a random regressor fixture.
    """
    X, y = make_regression(
        n_samples=500, n_features=22, n_informative=8, random_state=2019
    )

    # Set a class attribute for continuous data
    request.cls.continuous = Dataset(X, y)


@pytest.fixture(scope="class")
def s_curves(request):
    """
    Creates a random regressor fixture.
    """
    X, y = make_s_curve(1000, random_state=888)
    # Set a class attribute for continuous data
    request.cls.s_curves = Dataset(X, y)



# tests.test_model_selection.test_cross_validation
# Tests for the CVScores visualizer
#
# Author:  Rebecca Bilbro
# Created: Fri Aug 10 13:45:11 2018 -0400
#
# Copyright (C) 2018 The scikit-yb developers
# For license information, see LICENSE.txt
#
# ID: test_cross_validation.py [962c8bb] rebeccabilbro@users.noreply.github.com $

"""
Tests for the CVScores visualizer
"""

##########################################################################
# Imports
##########################################################################

import pytest
import numpy.testing as npt

from unittest.mock import patch
from tests.base import VisualTestCase

from sklearn.svm import SVC
from sklearn.naive_bayes import BernoulliNB
from sklearn.tree import DecisionTreeRegressor
from sklearn.preprocessing import OneHotEncoder, MinMaxScaler
from sklearn.neighbors import KNeighborsClassifier
from sklearn.linear_model import RidgeCV, LogisticRegressionCV
from sklearn.model_selection import ShuffleSplit, StratifiedKFold
from sklearn.pipeline import Pipeline

from yellowbrick.datasets import load_mushroom
from yellowbrick.model_selection.cross_validation import *


try:
    import pandas as pd
except ImportError:
    pd = None


##########################################################################
# Test Cases
##########################################################################


@pytest.mark.usefixtures("classification", "regression")
class TestCrossValidation(VisualTestCase):
    """
    Test the CVScores visualizer
    """

    @patch.object(CVScores, "draw")
    def test_fit(self, mock_draw):
        """
        Assert that fit returns self and creates expected properties
        """
        X, y = self.classification

        params = ("cv_scores_", "cv_scores_mean_")

        oz = CVScores(SVC())

        for param in params:
            assert not hasattr(oz, param)

        assert oz.fit(X, y) is oz
        mock_draw.assert_called_once()

        for param in params:
            assert hasattr(oz, param)

    def test_classifier(self):
        """
        Test image closeness on a classification dataset with kNN
        """
        X, y = self.classification

        cv = ShuffleSplit(3, random_state=288)

        oz = CVScores(KNeighborsClassifier(), cv=cv, scoring="f1_weighted")

        oz.fit(X, y)
        oz.finalize()

        self.assert_images_similar(oz, tol=2.0)

    def test_classifier_with_cv(self):
        """
        Test that CVScores passes cv param to internal sklearn classifier with CV
        """
        X, y = self.classification

        cv = StratifiedKFold(5, shuffle=True, random_state=288)
        oz_external_cv = CVScores(LogisticRegressionCV(), cv=cv)
        oz_internal_cv = CVScores(LogisticRegressionCV(cv=cv))
        oz_external_cv.fit(X, y)
        oz_internal_cv.fit(X, y)

        npt.assert_array_almost_equal(
            oz_external_cv.cv_scores_, oz_internal_cv.cv_scores_, decimal=1
        )

    def test_regression(self):
        """
        Test image closeness on a regression dataset with a DecisionTree
        """
        X, y = self.regression

        cv = ShuffleSplit(5, random_state=938)
        oz = CVScores(DecisionTreeRegressor(random_state=23), cv=cv, scoring="r2")
        oz.fit(X, y)
        oz.finalize()

        self.assert_images_similar(oz, tol=36.0)

    def test_regressor_with_cv(self):
        """
        Test that CVScores passes cv param to internal sklearn regressor with CV
        """
        X, y = self.regression

        cv = ShuffleSplit(5, random_state=288)
        oz_external_cv = CVScores(RidgeCV(), cv=cv)
        oz_internal_cv = CVScores(RidgeCV(cv=cv))
        oz_external_cv.fit(X, y)
        oz_internal_cv.fit(X, y)

        npt.assert_array_almost_equal(
            oz_external_cv.cv_scores_, oz_internal_cv.cv_scores_
        )

    def test_quick_method(self):
        """
        Test cross validation quick method with image closeness on SVC
        """
        X, y = self.classification

        cv = ShuffleSplit(n_splits=5, test_size=0.2, random_state=321)
        viz = cv_scores(SVC(), X, y, cv=cv, show=False)

        self.assert_images_similar(viz, tol=2.0)

    @pytest.mark.skipif(pd is None, reason="test requires pandas")
    def test_pandas_integration(self):
        """
        Test on mushroom dataset with pandas DataFrame and Series and NB
        """
        data = load_mushroom(return_dataset=True)
        X, y = data.to_pandas()

        X = pd.get_dummies(X)

        assert isinstance(X, pd.DataFrame)
        assert isinstance(y, pd.Series)

        cv = StratifiedKFold(n_splits=2, shuffle=True, random_state=11)
        oz = CVScores(BernoulliNB(), cv=cv)

        oz.fit(X, y)
        oz.finalize()

        self.assert_images_similar(oz, tol=2.0)

    def test_numpy_integration(self):
        """
        Test on mushroom dataset with NumPy arrays
        """
        data = load_mushroom(return_dataset=True)
        X, y = data.to_numpy()

        X = OneHotEncoder().fit_transform(X).toarray()

        cv = StratifiedKFold(n_splits=2, shuffle=True, random_state=11)
        oz = CVScores(BernoulliNB(), cv=cv)

        oz.fit(X, y)
        oz.finalize()

        self.assert_images_similar(oz, tol=2.0)

    def test_within_pipeline(self):
        """
        Test that visualizer can be accessed within a sklearn pipeline
        """
        X, y = load_mushroom(return_dataset=True).to_numpy()
        X = OneHotEncoder().fit_transform(X).toarray()

        cv = StratifiedKFold(n_splits=2, shuffle=True, random_state=11)
        model = Pipeline([
            ('minmax', MinMaxScaler()), 
            ('cvscores', CVScores(BernoulliNB(), cv=cv))
        ])

        model.fit(X, y)
        model['cvscores'].finalize()
        self.assert_images_similar(model['cvscores'], tol=2.0)

    def test_within_pipeline_quickmethod(self):
        """
        Test that visualizer quickmethod can be accessed within a
        sklearn pipeline
        """
        X, y = load_mushroom(return_dataset=True).to_numpy()
        X = OneHotEncoder().fit_transform(X).toarray()
        
        cv = StratifiedKFold(n_splits=2, shuffle=True, random_state=11)
        model = Pipeline([
            ('minmax', MinMaxScaler()), 
            ('cvscores', cv_scores(BernoulliNB(), X, y, cv=cv, show=False,
                                      random_state=42))
            ])
        self.assert_images_similar(model['cvscores'], tol=2.0)

    def test_pipeline_as_model_input(self):
        """
        Test that visualizer can handle sklearn pipeline as model input
        """
        X, y = load_mushroom(return_dataset=True).to_numpy()
        X = OneHotEncoder().fit_transform(X).toarray()

        cv = StratifiedKFold(n_splits=2, shuffle=True, random_state=11)
        model = Pipeline([
            ('minmax', MinMaxScaler()), 
            ('nb', BernoulliNB())
        ])

        oz = CVScores(model, cv=cv)
        oz.fit(X, y)
        oz.finalize()
        self.assert_images_similar(oz, tol=2.0)

    def test_pipeline_as_model_input_quickmethod(self):
        """
        Test that visualizer can handle sklearn pipeline as model input
        within a quickmethod
        """
        X, y = load_mushroom(return_dataset=True).to_numpy()
        X = OneHotEncoder().fit_transform(X).toarray()

        cv = StratifiedKFold(n_splits=2, shuffle=True, random_state=11)
        model = Pipeline([
            ('minmax', MinMaxScaler()), 
            ('nb', BernoulliNB())
        ])

        oz = cv_scores(model, X, y, show=False, cv=cv)
        self.assert_images_similar(oz, tol=2.0)

    def test_get_params(self):
        """
        Ensure dropping curve get params works correctly
        """
        oz = CVScores(BernoulliNB())
        params = oz.get_params()
        assert len(params) > 0


# tests.test_text.test_umap
# Tests for the UMAP visual corpus embedding mechanism.
#
# Author:   John Healy
# Created:  Mon Dec 03, 14:00:00
#
# Copyright (C) 2018 The scikit-yb developers
# For license information, see LICENSE.txt
#
# ID: test_umap.py [] jchealy@gmail.com> $

"""
Tests for the UMAP visual corpus embedding mechanism.
"""

##########################################################################
## Imports
##########################################################################

import sys
import pytest
import warnings

from unittest import mock
from tests.base import VisualTestCase
from yellowbrick.text.umap_vis import *
from yellowbrick.datasets import load_hobbies
from yellowbrick.exceptions import YellowbrickValueError

from sklearn.datasets import make_classification
from sklearn.feature_extraction.text import TfidfVectorizer

try:
    import pandas
except ImportError:
    pandas = None

try:
    from umap import UMAP
except ImportError:
    UMAP = None
except (RuntimeError, AttributeError):
    UMAP = None
    warnings.warn(
        "Error Importing UMAP.  UMAP does not support python 2.7 on Windows 32 bit."
    )


##########################################################################
## Data
##########################################################################

corpus = load_hobbies()


##########################################################################
## UMAP Tests
##########################################################################


@mock.patch("yellowbrick.text.umap_vis.UMAP", None)
def test_umap_unavailable():
    """
    Assert an appropriate exception is raised when UMAP is not installed
    """
    from yellowbrick.text.umap_vis import UMAP

    assert UMAP is None

    with pytest.raises(
        YellowbrickValueError, match="umap package doesn't seem to be installed"
    ):
        UMAPVisualizer()


@pytest.mark.skipif(UMAP is None, reason="tests require the umap library")
@pytest.mark.xfail(
    sys.platform == "win32", reason="not supported on windows 32bit with Python 2.7"
)
class TestUMAP(VisualTestCase):
    """
    UMAPVisualizer tests
    """

    def test_make_pipeline(self):
        """
        Verify the pipeline creation step for UMAP
        """

        umap = UMAPVisualizer()  # Should not cause an exception.
        assert umap.transformer_ is not None

        assert len(umap.transformer_.steps) == 1

    def test_integrated_umap(self):
        """
        Check UMAP integrated visualization on the hobbies corpus
        """
        tfidf = TfidfVectorizer()

        docs = tfidf.fit_transform(corpus.data)
        labels = corpus.target

        umap = UMAPVisualizer(random_state=8392, colormap="Set1", alpha=1.0)
        umap.fit_transform(docs, labels)

        tol = 55
        self.assert_images_similar(umap, tol=tol)

    def test_sklearn_umap_size(self):
        """
        Check to make sure sklearn's UMAP doesn't use the size param
        """
        # In UMAPVisualizer, the internal sklearn UMAP transform consumes
        # some but not all kwargs passed in by user. Those not in get_params(),
        # like size, are passed through to YB's finalize method. This test should
        # notify us  if UMAP's params change on the sklearn side.
        with pytest.raises(TypeError):
            UMAP(size=(100, 100))

    def test_sklearn_umap_title(self):
        """
        Check to make sure sklearn's UMAP doesn't use the title param
        """
        # In TSNEVisualizer, the internal sklearn UMAP transform consumes
        # some but not all kwargs passed in by user. Those not in get_params(),
        # like title, are passed through to YB's finalize method. This test should
        # notify us  if UMAP's params change on the sklearn side.
        with pytest.raises(TypeError):
            UMAP(title="custom_title")

    def test_custom_title_umap(self):
        """
        Check UMAP can accept a custom title (string) from the user
        """
        umap = UMAPVisualizer(title="custom_title")

        assert umap.title == "custom_title"

    def test_custom_size_umap(self):
        """
        Check UMAP can accept a custom size (tuple of pixels) from the user
        """
        umap = UMAPVisualizer(size=(100, 50))

        assert umap._size == (100, 50)

    def test_custom_colors_umap(self):
        """
        Check UMAP accepts and properly handles custom colors from user
        """
        ## produce random data
        X, y = make_classification(
            n_samples=200,
            n_features=100,
            n_informative=20,
            n_redundant=10,
            n_classes=5,
            random_state=42,
        )

        ## specify a list of custom colors >= n_classes
        purple_blues = ["indigo", "orchid", "plum", "navy", "purple", "blue"]

        ## instantiate the visualizer and check that self.colors is correct
        purple_umap = UMAPVisualizer(colors=purple_blues, random_state=87)
        assert purple_umap.colors == purple_blues

        ## fit the visualizer and check that self.color_values is as long as
        ## n_classes and is the first n_classes items in self.colors
        purple_umap.fit(X, y)
        assert len(purple_umap.color_values_) == len(purple_umap.classes_)
        assert purple_umap.color_values_ == purple_blues[: len(purple_umap.classes_)]

        ## specify a list of custom colors < n_classes
        greens = ["green", "lime", "teal"]

        ## instantiate the visualizer and check that self.colors is correct
        green_umap = UMAPVisualizer(colors=greens, random_state=87)
        assert green_umap.colors == greens

        ## fit the visualizer and check that self.color_values is as long as
        ## n_classes and the user-supplied color list gets recycled as expected
        green_umap.fit(X, y)
        assert len(green_umap.color_values_) == len(green_umap.classes_)
        assert green_umap.color_values_ == ["green", "lime", "teal", "green", "lime"]

    def test_make_classification_umap(self):
        """
        Test UMAP integrated visualization on a sklearn classifier dataset
        """

        ## produce random data
        X, y = make_classification(
            n_samples=200,
            n_features=100,
            n_informative=20,
            n_redundant=10,
            n_classes=3,
            random_state=42,
        )

        ## visualize data with UMAP
        umap = UMAPVisualizer(random_state=87)
        umap.fit(X, y)

        self.assert_images_similar(umap, tol=40)

    def test_make_classification_umap_class_labels(self):
        """
        Test UMAP integrated visualization with class labels specified
        """

        ## produce random data
        X, y = make_classification(
            n_samples=200,
            n_features=100,
            n_informative=20,
            n_redundant=10,
            n_classes=3,
            random_state=42,
        )

        ## visualize data with UMAP
        umap = UMAPVisualizer(random_state=87, labels=["a", "b", "c"])
        umap.fit(X, y)

        self.assert_images_similar(umap, tol=40)

    def test_umap_mismtached_labels(self):
        """
        Assert exception is raised when number of labels doesn't match
        """
        ## produce random data
        X, y = make_classification(
            n_samples=200,
            n_features=100,
            n_informative=20,
            n_redundant=10,
            n_classes=3,
            random_state=42,
        )

        ## fewer labels than classes
        umap = UMAPVisualizer(random_state=87, labels=["a", "b"])
        with pytest.raises(YellowbrickValueError):
            umap.fit(X, y)

        ## more labels than classes
        umap = UMAPVisualizer(random_state=87, labels=["a", "b", "c", "d"])
        with pytest.raises(YellowbrickValueError):
            umap.fit(X, y)

    def test_no_target_umap(self):
        """
        Test UMAP when no target or classes are specified
        """
        ## produce random data
        X, y = make_classification(
            n_samples=200,
            n_features=100,
            n_informative=20,
            n_redundant=10,
            n_classes=3,
            random_state=6897,
        )

        ## visualize data with UMAP
        umap = UMAPVisualizer(random_state=64)
        umap.fit(X)

        self.assert_images_similar(umap, tol=40)

    @pytest.mark.skipif(pandas is None, reason="test requires pandas")
    def test_visualizer_with_pandas(self):
        """
        Test UMAP when passed a pandas DataFrame and series
        """
        X, y = make_classification(
            n_samples=200,
            n_features=100,
            n_informative=20,
            n_redundant=10,
            n_classes=3,
            random_state=3020,
        )

        X = pandas.DataFrame(X)
        y = pandas.Series(y)

        umap = UMAPVisualizer(random_state=64)
        umap.fit(X, y)

        self.assert_images_similar(umap, tol=40)

    def test_alpha_param(self):
        """
        Test that the user can supply an alpha param on instantiation
        """
        ## produce random data
        X, y = make_classification(
            n_samples=200,
            n_features=100,
            n_informative=20,
            n_redundant=10,
            n_classes=3,
            random_state=42,
        )

        ## Instantiate a UMAPVisualizer, provide custom alpha
        umap = UMAPVisualizer(random_state=64, alpha=0.5)

        # Test param gets set correctly
        assert umap.alpha == 0.5

        # Mock ax and fit the visualizer
        umap.ax = mock.MagicMock(autospec=True)
        umap.fit(X, y)

        # Test that alpha was passed to internal matplotlib scatterplot
        _, scatter_kwargs = umap.ax.scatter.call_args
        assert "alpha" in scatter_kwargs
        assert scatter_kwargs["alpha"] == 0.5

    def test_quick_method(self):
        """
        Test for umap quick  method with hobbies dataset
        """
        corpus = load_hobbies()
        tfidf = TfidfVectorizer()

        X = tfidf.fit_transform(corpus.data)
        y = corpus.target

        viz = umap(X, y, show=False)
        assert isinstance(viz, UMAPVisualizer)

        self.assert_images_similar(viz, tol=50)




# tests.test_meta
# Meta testing for testing helper functions!
#
# Author:  Benjamin Bengfort <benjamin@bengfort.com>
# Created: Sat Apr 07 13:16:53 2018 -0400
#
# ID: test_meta.py [0a2d2b4] benjamin@bengfort.com $

"""
Meta testing for testing helper functions!
"""

##########################################################################
## Imports
##########################################################################

import os
import pytest
import inspect

import matplotlib as mpl

from tests.rand import RandomVisualizer
from unittest.mock import MagicMock, patch
from tests.base import ACTUAL_IMAGES, BASELINE_IMAGES
from tests.base import VisualTestCase, ImageComparison

from yellowbrick.exceptions import ImageComparisonFailure


def assert_path_exists(*parts):
    # Hide this method from the pytest traceback on test failure.
    __tracebackhide__ = True

    path = os.path.join(*parts)
    assert os.path.exists(path), "expected {} to exist".format(path)


def assert_path_not_exists(*parts):
    # Hide this method from the pytest traceback on test failure.
    __tracebackhide__ = True

    path = os.path.join(*parts)
    assert not os.path.exists(path), "expected {} not to exist".format(path)


##########################################################################
## Test Cases
##########################################################################


class TestMetaImageComparison(VisualTestCase):
    """
    Meta Test: ImageComparison test cases
    """

    def test_image_comparison(self):
        """
        Test the image comparison initialization and properties
        """

        def inner_assertion_function(ax):
            stack = inspect.stack()
            return ImageComparison(stack, ax=ax)

        ax = MagicMock()
        compare = inner_assertion_function(ax)
        assert compare.ax is ax
        assert compare.test_func_name == "test_image_comparison"
        assert compare.test_module_path == "test_meta"

        # Must use os.path.join for Windows/POSIX compatibility
        assert compare.actual_image_path.endswith(
            os.path.join(
                "tests", "actual_images", "test_meta", "test_image_comparison.png"
            )
        )
        assert compare.baseline_image_path.endswith(
            os.path.join(
                "tests", "baseline_images", "test_meta", "test_image_comparison.png"
            )
        )

    @patch.object(ImageComparison, "cleanup")
    @patch.object(ImageComparison, "save")
    @patch.object(ImageComparison, "compare")
    def test_image_comparison_call(self, mock_cleanup, mock_save, mock_compare):
        """
        Test that image comparison cleans up, saves, and compares
        """

        def inner_assertion_function():
            stack = inspect.stack()
            return ImageComparison(stack, ax=MagicMock())

        compare = inner_assertion_function()
        compare()

        mock_cleanup.assert_called_once()
        mock_save.assert_called_once()
        mock_compare.assert_called_once()

    def test_image_comparison_requires_ax(self):
        """
        Assert raises if an axes object is not supplied
        """
        with pytest.raises(ValueError, match="ax must be specified"):
            ImageComparison(inspect.stack())

    def test_image_comparison_not_in_assertion(self):
        """
        Assert raises when image comparison not in assertion helper
        """
        with pytest.raises(ValueError, match="not a test function"):
            stack = inspect.stack()
            ImageComparison(stack, ax=MagicMock())

    def test_missing_baseline_image(self):
        """
        Test that a missing baseline image raises an exception
        """
        viz = RandomVisualizer(random_state=14).fit()
        viz.finalize()

        # Assert the baseline image does not exist
        assert_path_not_exists(
            BASELINE_IMAGES, "test_meta", "test_missing_baseline_image.png"
        )

        with pytest.raises(ImageComparisonFailure, match="image does not exist"):
            self.assert_images_similar(viz)

        # Assert the actual image was created (to copy to baseline)
        assert_path_exists(
            ACTUAL_IMAGES, "test_meta", "test_missing_baseline_image.png"
        )

    def test_random_visualizer(self):
        """
        Test that a random visualization is correctly compared to a baseline
        """
        viz = RandomVisualizer(random_state=111).fit()
        viz.finalize()

        assert mpl.get_backend().lower() == "agg"

        compare = self.assert_images_similar(viz, tol=1.0)
        assert_path_exists(compare.actual_image_path)
        assert_path_exists(compare.baseline_image_path)

    def test_random_visualizer_not_close(self):
        """
        Test that not close visualizers raise an assertion error.
        """
        # Baseline image random_state=224
        # NOTE: if regenerating baseline images, skip this one or change random state!
        viz = RandomVisualizer(random_state=224).fit()
        viz.finalize()

        with pytest.raises(ImageComparisonFailure, match="images not close"):
            # If failing, perhaps baseline images were regenerated? See note above.
            self.assert_images_similar(viz, tol=0.5)

        # Assert there is a diff
        assert_path_exists(
            ACTUAL_IMAGES,
            "test_meta",
            "test_random_visualizer_not_close-failed-diff.png",
        )

    def test_random_visualizer_increased_tolerance(self):
        """
        Test that not close visualizers pass with increased tolerance
        """
        viz = RandomVisualizer(random_state=224).fit()
        viz.finalize()

        self.assert_images_similar(viz, tol=30)



# tests.test_text.test_tsne
# Tests for the TSNE visual corpus embedding mechanism.
#
# Author:   Benjamin Bengfort
# Created:  Mon Feb 20 07:23:53 2017 -0500
#
# Copyright (C) 2018 The scikit-yb developers
# For license information, see LICENSE.txt
#
# ID: test_tsne.py [6aa9198] benjamin@bengfort.com $

"""
Tests for the TSNE visual corpus embedding mechanism.
"""

##########################################################################
## Imports
##########################################################################

import pytest
from unittest import mock

from yellowbrick.text.tsne import *
from tests.base import VisualTestCase
from yellowbrick.datasets import load_hobbies
from yellowbrick.exceptions import YellowbrickValueError

from sklearn.manifold import TSNE
from sklearn.datasets import make_classification
from sklearn.feature_extraction.text import TfidfVectorizer

try:
    import pandas
except ImportError:
    pandas = None


##########################################################################
## Data
##########################################################################

corpus = load_hobbies()


##########################################################################
## TSNE Tests
##########################################################################


class TestTSNE(VisualTestCase):
    """
    TSNEVisualizer tests
    """

    def test_bad_decomposition(self):
        """
        Ensure an error is raised when a bad decompose argument is specified
        """
        with pytest.raises(YellowbrickValueError):
            TSNEVisualizer(decompose="bob")

    def test_make_pipeline(self):
        """
        Verify the pipeline creation step for TSNE
        """

        tsne = TSNEVisualizer()  # Should not cause an exception.
        assert tsne.transformer_ is not None

        svdp = tsne.make_transformer("svd", 90)
        assert len(svdp.steps) == 2

        pcap = tsne.make_transformer("pca")
        assert len(pcap.steps) == 2

        none = tsne.make_transformer(None)
        assert len(none.steps) == 1

    def test_integrated_tsne(self):
        """
        Check tSNE integrated visualization on the hobbies corpus
        """
        tfidf = TfidfVectorizer()

        docs = tfidf.fit_transform(corpus.data)
        labels = corpus.target

        tsne = TSNEVisualizer(random_state=8392, colormap="Set1", alpha=1.0)
        tsne.fit_transform(docs, labels)

        self.assert_images_similar(tsne, tol=50)

    def test_sklearn_tsne_size(self):
        """
        Check to make sure sklearn's TSNE doesn't use the size param
        """
        # In TSNEVisualizer, the internal sklearn TSNE transform consumes
        # some but not all kwargs passed in by user. Those not in get_params(),
        # like size, are passed through to YB's finalize method. This test should
        # notify us  if TSNE's params change on the sklearn side.
        with pytest.raises(TypeError):
            TSNE(size=(100, 100))

    def test_sklearn_tsne_title(self):
        """
        Check to make sure sklearn's TSNE doesn't use the title param
        """
        # In TSNEVisualizer, the internal sklearn TSNE transform consumes
        # some but not all kwargs passed in by user. Those not in get_params(),
        # like title, are passed through to YB's finalize method. This test should
        # notify us  if TSNE's params change on the sklearn side.
        with pytest.raises(TypeError):
            TSNE(title="custom_title")

    def test_custom_title_tsne(self):
        """
        Check tSNE can accept a custom title (string) from the user
        """
        tsne = TSNEVisualizer(title="custom_title")

        assert tsne.title == "custom_title"

    def test_custom_size_tsne(self):
        """
        Check tSNE can accept a custom size (tuple of pixels) from the user
        """
        tsne = TSNEVisualizer(size=(100, 50))

        assert tsne._size == (100, 50)

    def test_custom_colors_tsne(self):
        """
        Check tSNE accepts and properly handles custom colors from user
        """
        ## produce random data
        X, y = make_classification(
            n_samples=200,
            n_features=100,
            n_informative=20,
            n_redundant=10,
            n_classes=5,
            random_state=42,
        )

        ## specify a list of custom colors >= n_classes
        purple_blues = ["indigo", "orchid", "plum", "navy", "purple", "blue"]

        ## instantiate the visualizer and check that self.colors is correct
        purple_tsne = TSNEVisualizer(colors=purple_blues, random_state=87)
        assert purple_tsne.colors == purple_blues

        ## fit the visualizer and check that self.color_values is as long as
        ## n_classes and is the first n_classes items in self.colors
        purple_tsne.fit(X, y)
        assert len(purple_tsne.color_values_) == len(purple_tsne.classes_)
        assert purple_tsne.color_values_ == purple_blues[: len(purple_tsne.classes_)]

        ## specify a list of custom colors < n_classes
        greens = ["green", "lime", "teal"]

        ## instantiate the visualizer and check that self.colors is correct
        green_tsne = TSNEVisualizer(colors=greens, random_state=87)
        assert green_tsne.colors == greens

        ## fit the visualizer and check that self.color_values is as long as
        ## n_classes and the user-supplied color list gets recycled as expected
        green_tsne.fit(X, y)
        assert len(green_tsne.color_values_) == len(green_tsne.classes_)
        assert green_tsne.color_values_ == ["green", "lime", "teal", "green", "lime"]

    def test_make_classification_tsne(self):
        """
        Test tSNE integrated visualization on a sklearn classifier dataset
        """

        ## produce random data
        X, y = make_classification(
            n_samples=200,
            n_features=100,
            n_informative=20,
            n_redundant=10,
            n_classes=3,
            random_state=42,
        )

        ## visualize data with t-SNE
        tsne = TSNEVisualizer(random_state=87)
        tsne.fit(X, y)

        self.assert_images_similar(tsne, tol=35)

    def test_make_classification_tsne_class_labels(self):
        """
        Test tSNE integrated visualization with class labels specified
        """

        ## produce random data
        X, y = make_classification(
            n_samples=200,
            n_features=100,
            n_informative=20,
            n_redundant=10,
            n_classes=3,
            random_state=42,
        )

        ## visualize data with t-SNE
        tsne = TSNEVisualizer(random_state=87, labels=["a", "b", "c"])
        tsne.fit(X, y)

        self.assert_images_similar(tsne, tol=35)

    def test_tsne_mismtached_labels(self):
        """
        Assert exception is raised when number of labels doesn't match
        """
        ## produce random data
        X, y = make_classification(
            n_samples=200,
            n_features=100,
            n_informative=20,
            n_redundant=10,
            n_classes=3,
            random_state=42,
        )

        ## fewer labels than classes
        tsne = TSNEVisualizer(random_state=87, labels=["a", "b"])
        with pytest.raises(YellowbrickValueError):
            tsne.fit(X, y)

        ## more labels than classes
        tsne = TSNEVisualizer(random_state=87, labels=["a", "b", "c", "d"])
        with pytest.raises(YellowbrickValueError):
            tsne.fit(X, y)

    def test_no_target_tsne(self):
        """
        Test tSNE when no target or classes are specified
        """
        ## produce random data
        X, y = make_classification(
            n_samples=200,
            n_features=100,
            n_informative=20,
            n_redundant=10,
            n_classes=3,
            random_state=6897,
        )

        ## visualize data with t-SNE
        tsne = TSNEVisualizer(random_state=64)
        tsne.fit(X)

        self.assert_images_similar(tsne, tol=35)

    @pytest.mark.skipif(pandas is None, reason="test requires pandas")
    def test_visualizer_with_pandas(self):
        """
        Test tSNE when passed a pandas DataFrame and series
        """
        X, y = make_classification(
            n_samples=200,
            n_features=100,
            n_informative=20,
            n_redundant=10,
            n_classes=3,
            random_state=3020,
        )

        X = pandas.DataFrame(X)
        y = pandas.Series(y)

        tsne = TSNEVisualizer(random_state=64)
        tsne.fit(X, y)

        self.assert_images_similar(tsne, tol=35)

    def test_alpha_param(self):
        """
        Test that the user can supply an alpha param on instantiation
        """
        ## produce random data
        X, y = make_classification(
            n_samples=200,
            n_features=100,
            n_informative=20,
            n_redundant=10,
            n_classes=3,
            random_state=42,
        )

        ## Instantiate a TSNEVisualizer, provide custom alpha
        tsne = TSNEVisualizer(random_state=64, alpha=0.5)

        # Test param gets set correctly
        assert tsne.alpha == 0.5

        # Mock ax and fit the visualizer
        tsne.ax = mock.MagicMock(autospec=True)
        tsne.fit(X, y)

        # Test that alpha was passed to internal matplotlib scatterplot
        _, scatter_kwargs = tsne.ax.scatter.call_args
        assert "alpha" in scatter_kwargs
        assert scatter_kwargs["alpha"] == 0.5

    def test_quick_method(self):
        """
        Test for tsne quick  method with hobbies dataset
        """
        corpus = load_hobbies()
        tfidf = TfidfVectorizer()

        X = tfidf.fit_transform(corpus.data)
        y = corpus.target

        viz = tsne(X, y, show=False)

        self.assert_images_similar(viz, tol=50)



# tests.test_utils.test_wrapper
# Testing for the wrapping utility.
#
# Author:   Benjamin Bengfort <bbengfort@districtdatalabs.com>
# Created:  Mon May 22 09:25:52 2017 -0700
#
# Copyright (C) 2017 The scikit-yb developers
# For license information, see LICENSE.txt
#
# ID: test_wrapper.py [b2ecd50] benjamin@bengfort.com $

"""
Testing for the wrapping utility.
"""

##########################################################################
## Imports
##########################################################################

import pytest

from unittest import mock

from yellowbrick.base import Visualizer
from yellowbrick.utils.wrapper import *
from yellowbrick.exceptions import YellowbrickAttributeError, YellowbrickTypeError
from sklearn.naive_bayes import MultinomialNB
from sklearn.naive_bayes import GaussianNB


##########################################################################
## Fixture
##########################################################################


class MockVisualizer(Visualizer):
    def __init__(self, ax=None, **kwargs):
        self.ax = ax
        self.fit = mock.MagicMock()
        self.finalize = mock.MagicMock()
        self.show = mock.MagicMock()
        self.set_title = mock.MagicMock()

    @property
    def ax(self):
        return self._ax

    @ax.setter
    def ax(self, val):
        self._ax = val


class WrappedEstimator(MockVisualizer, Wrapper):
    def __init__(self, **kwargs):
        self.estimator = mock.MagicMock(spec=MultinomialNB())

        Wrapper.__init__(self, self.estimator)
        MockVisualizer.__init__(self, **kwargs)

    def draw(self):
        return True

    def foo(self, a, b):
        return a + b


##########################################################################
## Wrapper Test Case
##########################################################################


class TestWrapper(object):
    """
    Test the object Wrapper mixin utility
    """

    def test_wrapper_methods(self):
        """
        Asssert that local wrapper methods are called
        """
        obj = WrappedEstimator()

        # Assert that all the wrapper methods are called
        assert obj.draw()
        assert obj.foo(2, 2) == 4
        assert obj.estimator is not None

    def test_super_methods(self):
        """
        Assert that Visualizer super methods are called
        """

        obj = WrappedEstimator()

        # Assert that visualizer methods are called
        obj.fit()
        obj.finalize()
        obj.show()
        obj.set_title()

        assert obj.ax is None
        obj.fit.assert_called_once_with()
        obj.finalize.assert_called_once_with()
        obj.show.assert_called_once_with()
        obj.set_title.assert_called_once_with()

    def test_wrapped_methods(self):
        """
        Assert that wrapped estimator methods are called
        """
        obj = WrappedEstimator()

        # Assert that estimator methods are called
        obj.predict()
        obj.predict_proba()
        obj.score()

        obj._wrapped.predict.assert_called_once_with()
        obj._wrapped.predict_proba.assert_called_once_with()
        obj._wrapped.score.assert_called_once_with()

    def test_rewrap_object(self):
        """
        Test the ability to "rewrap" an object on demand
        """
        obj = WrappedEstimator()
        old = obj._wrapped
        new = mock.MagicMock(spec=GaussianNB())

        obj.predict()
        old.predict.assert_called_once()
        new.assert_not_called()

        # rewrap
        obj._wrapped = new
        obj.predict()
        old.predict.assert_called_once()
        new.predict.assert_called_once()

    def test_wrapper_recursion(self):
        """
        Ensure wrapper recursion isn't possible
        """
        obj = Wrapper("")
        obj._wrapped = obj
        with pytest.raises(YellowbrickTypeError):
            obj.foo

    def test_attribute_error(self):
        """
        Attribute errors should return a YellowbrickAttributeError
        """
        obj = WrappedEstimator()
        pat = r"neither visualizer 'WrappedEstimator' nor wrapped estimator 'MagicMock' have attribute 'notaproperty'"
        with pytest.raises(YellowbrickAttributeError, match=pat):
            obj.notaproperty


# tests.test_text.test_dispersion
# Tests for the dispersion plot visualization
#
# Author:   Larry Gray
# Created:  2018-06-22 15:27
#
# Copyright (C) 2018 The scikit-yb developers
# For license information, see LICENSE.txt
#
# ID: test_dispersion.py [25f1b9a] lwgray@gmail.com $

"""
Tests for the dispersion plot text visualization
"""

##########################################################################
## Imports
##########################################################################

import pytest
import matplotlib.pyplot as plt

from yellowbrick.exceptions import YellowbrickValueError
from yellowbrick.datasets import load_hobbies
from yellowbrick.text.dispersion import DispersionPlot, dispersion
from tests.base import VisualTestCase


##########################################################################
## Data
##########################################################################

corpus = load_hobbies()

##########################################################################
## DispersionPlot Tests
##########################################################################


class TestDispersionPlot(VisualTestCase):
    def test_quick_method(self):
        """
        Assert no errors occur when using the quick method
        """
        _, ax = plt.subplots()

        text = [doc.split() for doc in corpus.data]
        search_terms = ["Game", "player", "score", "oil", "Man"]

        viz = dispersion(search_terms=search_terms, corpus=text, ax=ax, show=False)
        viz.ax.grid(False)

        self.assert_images_similar(viz, tol=25)

    def test_integrated_dispersion_plot(self):
        """
        Assert no errors occur during DispersionPlot integration
        """
        text = [doc.split() for doc in corpus.data]
        search_terms = ["Game", "player", "score", "oil", "Man"]

        visualizer = DispersionPlot(search_terms)
        visualizer.fit(text)
        visualizer.ax.grid(False)

        self.assert_images_similar(visualizer, tol=25)

    def test_dispersion_plot_ignore_case(self):
        """
        Assert no errors occur during DispersionPlot integration
        with ignore_case parameter turned on
        """
        text = [doc.split() for doc in corpus.data]
        search_terms = ["Game", "player", "score", "oil", "Man"]

        visualizer = DispersionPlot(search_terms, ignore_case=True)
        visualizer.fit(text)
        visualizer.ax.grid(False)

        self.assert_images_similar(visualizer, tol=25.5)

    def test_dispersion_plot_generator_input(self):
        """
        Assert no errors occur during dispersionPlot integration
        when the corpus' text type is a generator
        """
        text = [doc.split() for doc in corpus.data]
        search_terms = ["Game", "player", "score", "oil", "Man"]

        visualizer = DispersionPlot(search_terms, ignore_case=True)
        visualizer.fit(text)
        visualizer.ax.grid(False)

        self.assert_images_similar(visualizer, tol=25.5)

    def test_dispersion_plot_annotate_docs(self):
        """
        Assert no errors occur during DispersionPlot integration
        with annotate_docs parameter turned on
        """
        text = [doc.split() for doc in corpus.data]
        search_terms = ["girl", "she", "boy", "he", "man"]

        visualizer = DispersionPlot(search_terms, annotate_docs=True)
        visualizer.fit(text)
        visualizer.ax.grid(False)

        self.assert_images_similar(visualizer, tol=25.5)

    def test_dispersion_plot_color_by_class(self):
        """
        Assert no errors occur during DispersionPlot integration
        when target values are specified
        """
        target = corpus.target
        text = [doc.split() for doc in corpus.data]
        search_terms = ["girl", "she", "boy", "he", "man"]

        visualizer = DispersionPlot(search_terms)
        visualizer.fit(text, target)
        visualizer.ax.grid(False)

        self.assert_images_similar(visualizer, tol=25)

    def test_dispersion_plot_mismatched_labels(self):
        """
        Assert exception is raised when number of labels doesn't match
        """
        target = corpus.target
        text = [doc.split() for doc in corpus.data]
        search_terms = ["girl", "she", "boy", "he", "man"]

        visualizer = DispersionPlot(search_terms, annotate_docs=True, labels=["a", "b"])

        msg = (
            r"number of supplied labels \(\d\) "
            r"does not match the number of classes \(\d\)"
        )

        with pytest.raises(YellowbrickValueError, match=msg):
            visualizer.fit(text, target)



# tests.test_features.test_manifold
# Tests for the Manifold High Dimensional Visualizations
#
# Author:  Benjamin Bengfort <benjamin@bengfort.com>
# Created: Sat May 12 11:24:41 2018 -0400
#
# Copyright (C) 2018 The scikit-yb developers
# For license information, see LICENSE.txt
#
# ID: test_manifold.py [02f8c27] benjamin@bengfort.com $

"""
Tests for the Manifold High Dimensional Visualizations
"""

##########################################################################
## Imports
##########################################################################

import pytest

from yellowbrick.features.manifold import *
from yellowbrick.utils.types import is_estimator
from yellowbrick.features.base import TargetType
from yellowbrick.exceptions import YellowbrickValueError, ModelError, NotFitted

from sklearn.pipeline import Pipeline
from sklearn.decomposition import PCA
from sklearn.datasets import make_blobs
from sklearn.datasets import make_s_curve
from sklearn.manifold import LocallyLinearEmbedding

from unittest.mock import patch
from tests.base import VisualTestCase

try:
    import pandas as pd
except ImportError:
    pd = None


##########################################################################
## Manifold Visualizer Tests
##########################################################################


@pytest.mark.usefixtures("s_curves", "discrete", "continuous")
class TestManifold(VisualTestCase):
    """
    Test Manifold visualizer
    """

    @pytest.mark.parametrize(
        "algorithm",
        ["lle", "ltsa", "hessian", "modified", "isomap", "mds", "spectral", "tsne"],
    )
    def test_manifold_construction(self, algorithm):
        """
        Should be able to construct a manifold estimator from a string
        """
        message = "case failed for {}".format(algorithm)
        params = {"n_neighbors": 18, "random_state": 53}
        oz = Manifold(manifold=algorithm, **params)
        assert is_estimator(oz.manifold), message
        assert oz.manifold.get_params()["n_components"] == 2, message

        manifold_params = oz.manifold.get_params()
        for param, value in params.items():
            if param in manifold_params:
                assert value == manifold_params[param], message

    @pytest.mark.parametrize(
        "algorithm", ["lle", "ltsa", "hessian", "modified", "isomap", "spectral"]
    )
    def test_manifold_warning(self, algorithm):
        """
        Should raise a warning if n_neighbors not specified
        """
        message = "case failed for {}".format(algorithm)
        n_neighbors = 6 if algorithm == "hessian" else 5

        with pytest.warns(YellowbrickWarning):
            oz = Manifold(manifold=algorithm)
            assert oz.n_neighbors == n_neighbors, message

    @pytest.mark.parametrize("algorithm", ["mds", "tsne"])
    def test_manifold_no_warning(self, algorithm):
        """
        Should not raise a warning if n_neighbors not specified
        """
        import warnings

        message = "case failed for {}".format(algorithm)

        with warnings.catch_warnings():
            warnings.simplefilter("error")  # Turn warnings into errors
            # If a warning is raised, it will cause an error and fail the test
            try:
                Manifold(manifold=algorithm)
            except Warning as w:
                pytest.fail(f"{message}: {w}")

    def test_bad_manifold_exception(self):
        """
        Should raise a ValueError when a bad manifold is passed in
        """
        with pytest.raises(YellowbrickValueError, match="could not create manifold"):
            Manifold(manifold=32)

    def test_manifold_instance_construction(self):
        """
        Should allow a sklearn.Estimator object to be set as manifold
        """
        manifold = Pipeline(
            [
                ("pca", PCA(n_components=50)),
                ("lle", LocallyLinearEmbedding(n_components=2)),
            ]
        )

        oz = Manifold(manifold=manifold)
        assert oz.manifold is manifold

    @pytest.mark.filterwarnings("ignore:Conversion of the second argument")
    @pytest.mark.parametrize(
        "algorithm", ["lle", "ltsa", "hessian", "modified", "isomap"]
    )
    def test_manifold_algorithm_transform_fit(self, algorithm):
        """
        Test manifold fit with algorithms having transform implemented
        """
        X, y = make_s_curve(1000, random_state=94)
        with pytest.warns(YellowbrickWarning):
            manifold = Manifold(manifold=algorithm, target="auto")

        assert manifold.fit(X, y) is manifold, "fit did not return self"

    @pytest.mark.filterwarnings("ignore:Conversion of the second argument")
    @pytest.mark.parametrize("algorithm", ["mds", "spectral", "tsne"])
    def test_manifold_algorithm_no_transform_fit(self, algorithm):
        """
        Test manifold fit with algorithms not having transform implemented
        """
        X, y = self.s_curves
        msg = "requires data to be simultaneously fit and transformed"
        oz = Manifold(manifold=algorithm, n_neighbors=10, random_state=223)
        with pytest.raises(ModelError, match=msg):
            oz.fit(X)

    @patch("yellowbrick.features.manifold.Manifold.draw", spec=True)
    @pytest.mark.parametrize("projection", [2, 3])
    def test_manifold_fit_transform(self, mock_draw, projection):
        """
        Test manifold fit_transform method
        """
        X, y = self.s_curves
        manifold = Manifold(target="auto", projection=projection)

        assert not hasattr(manifold, "fit_time_")

        Xp = manifold.fit_transform(X, y)
        assert Xp.shape == (X.shape[0], projection)

        mock_draw.assert_called_once()
        assert hasattr(manifold, "fit_time_")
        assert manifold._target_color_type == TargetType.CONTINUOUS

    @patch("yellowbrick.features.manifold.Manifold.fit_transform", spec=True)
    @patch("yellowbrick.features.manifold.Manifold.draw", spec=True)
    @pytest.mark.parametrize("projection", [2, 3])
    def test_manifold_transform(self, mock_draw, mock_fit_transform, projection):
        """
        Test manifold transform method
        """
        X, y = self.s_curves
        manifold = Manifold(
            manifold="lle", target="auto", n_neighbors=5, projection=projection
        )

        manifold.fit(X, y)
        Xp = manifold.transform(X, y)
        assert Xp.shape == (X.shape[0], projection)

        mock_draw.assert_called_once()

    def test_manifold_no_transform(self):
        """
        Test the exception when manifold doesn't implement transform.
        """
        X, _ = self.s_curves
        manifold = Manifold(manifold="lle", n_neighbors=5, target="auto")

        msg = "instance is not fitted yet, please call fit"
        with pytest.raises(NotFitted, match=msg):
            manifold.transform(X)

    @patch("yellowbrick.features.manifold.Manifold.fit", spec=True)
    @pytest.mark.parametrize("manifolds", ["mds", "spectral", "tsne"])
    def test_manifold_assert_no_transform(self, mock_fit, manifolds):
        """
        Assert that transform raises error when MDS, TSNE or
        Spectral Embedding algorithms are used.
        """
        X, _ = self.s_curves
        manifold = Manifold(manifold=manifolds, target="auto", n_neighbors=10)
        mock_fit(X)
        msg = "requires data to be simultaneously fit and transformed"
        with pytest.raises(ModelError, match=msg):
            manifold.transform(X)

    @pytest.mark.filterwarnings("ignore:Conversion of the second argument")
    def test_manifold_classification(self):
        """
        Image similarity test for classification dataset (discrete y)
        """
        X, y = self.discrete

        oz = Manifold(
            manifold="spectral", target="discrete", n_neighbors=5, random_state=108
        )
        assert not hasattr(oz, "classes_")

        oz.fit_transform(X, y)

        assert hasattr(oz, "classes_")
        assert not hasattr(oz, "range_")
        self.assert_images_similar(oz, tol=0.5)

    def test_manifold_classification_3d(self):
        """
        Image similarity test for classification dataset (discrete y)
        """
        X, y = self.discrete

        oz = Manifold(
            manifold="spectral",
            target="discrete",
            n_neighbors=5,
            random_state=108,
            projection=3,
        )

        assert not hasattr(oz, "classes_")

        oz.fit_transform(X, y)

        assert hasattr(oz, "classes_")
        assert not hasattr(oz, "range_")
        self.assert_images_similar(oz)

    def test_manifold_regression(self):
        """
        Image similarity test for regression dataset (continuous y)
        """
        X, y = self.continuous

        oz = Manifold(manifold="tsne", target="continuous", random_state=1)
        assert not hasattr(oz, "range_")

        oz.fit_transform(X, y)
        oz.finalize()
        assert not hasattr(oz, "classes_")
        assert hasattr(oz, "range_")
        self.assert_images_similar(oz, tol=23.1)

    def test_manifold_regression_3d(self):
        """
        Image similarity test for regression dataset (continuous y)
        """
        X, y = self.continuous

        oz = Manifold(
            manifold="tsne", target="continuous", random_state=1, projection=3
        )
        assert not hasattr(oz, "range_")

        oz.fit_transform(X, y)
        oz.finalize()
        oz.cbar.set_ticks([])
        assert not hasattr(oz, "classes_")
        assert hasattr(oz, "range_")
        self.assert_images_similar(oz, tol=15)

    def test_manifold_single(self):
        """
        Image similarity test for simple dataset (no y)
        """
        X, _ = make_blobs(n_samples=300, n_features=7, centers=3, random_state=1112)

        oz = Manifold(manifold="mds", random_state=139973)
        oz.fit_transform(X)

        self.assert_images_similar(oz, tol=44.8)

    def test_manifold_single_3d(self):
        """
        Image similarity test for simple dataset (no y)
        """
        X, _ = make_blobs(n_samples=300, n_features=7, centers=3, random_state=1112)

        oz = Manifold(manifold="mds", random_state=139973, projection=3)
        oz.fit_transform(X)

        self.assert_images_similar(oz, tol=14.3)

    @pytest.mark.skipif(pd is None, reason="requires pandas")
    def test_manifold_pandas(self):
        """
        Test manifold on a dataset made up of a pandas DataFrame and Series
        """
        X, y = self.s_curves

        oz = Manifold(
            manifold="ltsa",
            colormap="nipy_spectral",
            n_neighbors=10,
            target="continuous",
            random_state=223,
        )
        oz.fit_transform(X, y)
        oz.finalize()
        oz.cbar.set_ticks([])
        # TODO: find a way to decrease this tolerance
        self.assert_images_similar(oz, tol=40)

    def test_manifold_quick_method_no_target(self):
        """
        Test Manifold quick method with no target.
        """
        X, _ = make_blobs(n_samples=300, n_features=7, centers=3, random_state=1112)
        visualizer = manifold_embedding(
            X, manifold="mds", random_state=139973, show=False
        )

        assert isinstance(visualizer, Manifold)
        self.assert_images_similar(visualizer, tol=44.8)

    def test_manifold_quick_method_discrete_target(self):
        """
        Test Manifold quick method with a discrete target.
        """
        X, y = self.discrete

        visualizer = manifold_embedding(
            X,
            y,
            manifold="mds",
            target="discrete",
            n_neighbors=5,
            random_state=37,
            show=False,
        )
        assert isinstance(visualizer, Manifold)
        self.assert_images_similar(visualizer, tol=31.7)

    def test_manifold_quick_method_continuous_target(self):
        """
        Test Manifold quick method with a continuous target.
        """
        X, y = self.continuous

        visualizer = manifold_embedding(
            X, y, manifold="tsne", target="continuous", random_state=37, show=False
        )
        assert isinstance(visualizer, Manifold)

        # ImageComparisonFailure: images not close (RMS 18.879) across platforms
        self.assert_images_similar(visualizer, tol=22.7)



#!/usr/bin/env python
# Generate the classification report images for the tutorial

import matplotlib.pyplot as plt

from sklearn.pipeline import Pipeline
from sklearn.svm import LinearSVC, NuSVC, SVC
from sklearn.neighbors import KNeighborsClassifier
from sklearn.preprocessing import OneHotEncoder, LabelEncoder
from sklearn.linear_model import LogisticRegressionCV, LogisticRegression, SGDClassifier
from sklearn.ensemble import (
    BaggingClassifier,
    ExtraTreesClassifier,
    RandomForestClassifier,
)

from yellowbrick.datasets import load_mushroom
from yellowbrick.classifier import ClassificationReport


ESTIMATORS = {
    "SVC": {"model": SVC(gamma="auto"), "path": "images/tutorial/modelselect_svc.png"},
    "NuSVC": {
        "model": NuSVC(gamma="auto"),
        "path": "images/tutorial/modelselect_nu_svc.png",
    },
    "LinearSVC": {
        "model": LinearSVC(),
        "path": "images/tutorial/modelselect_linear_svc.png",
    },
    "SGD": {
        "model": SGDClassifier(max_iter=100, tol=1e-3),
        "path": "images/tutorial/modelselect_sgd_classifier.png",
    },
    "KNN": {
        "model": KNeighborsClassifier(),
        "path": "images/tutorial/modelselect_kneighbors_classifier.png",
    },
    "LR": {
        "model": LogisticRegression(solver="lbfgs"),
        "path": "images/tutorial/modelselect_logistic_regression.png",
    },
    "LRCV": {
        "model": LogisticRegressionCV(cv=3),
        "path": "images/tutorial/modelselect_logistic_regression_cv.png",
    },
    "Bags": {
        "model": BaggingClassifier(),
        "path": "images/tutorial/modelselect_bagging_classifier.png",
    },
    "XTrees": {
        "model": ExtraTreesClassifier(n_estimators=100),
        "path": "images/tutorial/modelselect_extra_trees_classifier.png",
    },
    "RF": {
        "model": RandomForestClassifier(n_estimators=100),
        "path": "images/tutorial/modelselect_random_forest_classifier.png",
    },
}


def visualize_model(X, y, estimator, path, **kwargs):
    """
    Test various estimators.
    """
    y = LabelEncoder().fit_transform(y)
    model = Pipeline([("one_hot_encoder", OneHotEncoder()), ("estimator", estimator)])

    _, ax = plt.subplots()

    # Instantiate the classification model and visualizer
    visualizer = ClassificationReport(
        model,
        classes=["edible", "poisonous"],
        cmap="YlGn",
        size=(600, 360),
        ax=ax,
        **kwargs
    )
    visualizer.fit(X, y)
    visualizer.score(X, y)
    visualizer.show(outpath=path)


if __name__ == "__main__":
    X, y = load_mushroom()

    for clf in ESTIMATORS.values():
        visualize_model(X, y, clf["model"], clf["path"])



# tests.test_regressor.test_prediction_error
# Ensure that the regressor prediction error visualization works.
#
# Author:   Rebecca Bilbro
# Author:   Benjamin Bengfort
# Created:  Sat Oct 8 16:30:39 2016 -0400
#
# Copyright (C) 2016 The scikit-yb developers
# For license information, see LICENSE.txt
#
# ID: test_prediction_error.py []  $

"""
Ensure that the regressor prediction error visualization works.
"""

##########################################################################
## Imports
##########################################################################

import pytest
import matplotlib.pyplot as plt
import numpy as np

from unittest import mock
from tests.fixtures import Dataset, Split
from tests.base import IS_WINDOWS_OR_CONDA, VisualTestCase

from yellowbrick.datasets import load_energy, load_concrete
from yellowbrick.regressor.prediction_error import PredictionError, prediction_error

from sklearn.datasets import make_regression
from sklearn.linear_model import Ridge, Lasso
from sklearn.neural_network import MLPRegressor
from sklearn.linear_model import LinearRegression
from sklearn.model_selection import train_test_split as tts
from sklearn.pipeline import Pipeline
from sklearn.impute import SimpleImputer

try:
    import pandas as pd
except ImportError:
    pd = None

##########################################################################
## Data
##########################################################################


@pytest.fixture(scope="class")
def data(request):
    """
    Creates a fixture of train and test splits for the sklearn digits dataset
    For ease of use returns a Dataset named tuple composed of two Split tuples.
    """
    X, y = make_regression(
        n_samples=500,
        n_features=22,
        n_informative=8,
        random_state=42,
        noise=0.2,
        bias=0.2,
    )

    X_train, X_test, y_train, y_test = tts(X, y, test_size=0.2, random_state=11)

    # Set a class attribute for digits
    request.cls.data = Dataset(Split(X_train, X_test), Split(y_train, y_test))


##########################################################################
## Prediction Error Test Cases
##########################################################################


@pytest.mark.usefixtures("data")
class TestPredictionError(VisualTestCase):
    """
    Test the PredictionError visualizer
    """

    @pytest.mark.filterwarnings("ignore:Stochastic Optimizer")
    @pytest.mark.filterwarnings("ignore:internal gelsd driver lwork query error")
    def test_prediction_error(self):
        """
        Test image similarity of prediction error on random data
        """
        _, ax = plt.subplots()

        model = MLPRegressor(random_state=229)
        visualizer = PredictionError(model, ax=ax)

        visualizer.fit(self.data.X.train, self.data.y.train)
        visualizer.score(self.data.X.test, self.data.y.test)
        visualizer.finalize()

        self.assert_images_similar(visualizer, tol=1, remove_legend=True)

    @pytest.mark.skipif(pd is None, reason="pandas is required")
    def test_prediction_error_pandas(self):
        """
        Test Pandas real world dataset with image similarity on Ridge
        """
        _, ax = plt.subplots()

        # Load the occupancy dataset from fixtures
        data = load_energy(return_dataset=True)
        X, y = data.to_pandas()

        # Create train/test splits
        splits = tts(X, y, test_size=0.2, random_state=8873)
        X_train, X_test, y_train, y_test = splits

        visualizer = PredictionError(Ridge(random_state=22), ax=ax)
        visualizer.fit(X_train, y_train)
        visualizer.score(X_test, y_test)
        visualizer.finalize()

        self.assert_images_similar(visualizer, tol=1, remove_legend=True)

    def test_prediction_error_numpy(self):
        """
        Test NumPy real world dataset with image similarity on Ridge
        """
        _, ax = plt.subplots()

        # Load the occupancy dataset from fixtures
        data = load_energy(return_dataset=True)
        X, y = data.to_numpy()

        # Create train/test splits
        splits = tts(X, y, test_size=0.2, random_state=8873)
        X_train, X_test, y_train, y_test = splits

        visualizer = PredictionError(Ridge(random_state=22), ax=ax)
        visualizer.fit(X_train, y_train)
        visualizer.score(X_test, y_test)
        visualizer.finalize()

        self.assert_images_similar(visualizer, tol=1, remove_legend=True)

    def test_score(self):
        """
        Assert returns R2 score
        """
        visualizer = PredictionError(LinearRegression())

        visualizer.fit(self.data.X.train, self.data.y.train)
        score = visualizer.score(self.data.X.test, self.data.y.test)

        assert score == pytest.approx(0.9999983124154965)
        assert visualizer.score_ == score

    def test_peplot_shared_limits(self):
        """
        Test shared limits on the peplot
        """
        visualizer = PredictionError(LinearRegression(), shared_limits=False)

        visualizer.fit(self.data.X.train, self.data.y.train)
        visualizer.score(self.data.X.test, self.data.y.test)
        visualizer.finalize()

        xlim = tuple(map(int, visualizer.ax.get_xlim()))
        ylim = tuple(map(int, visualizer.ax.get_ylim()))
        assert xlim == ylim

    @pytest.mark.filterwarnings("ignore:internal gelsd driver lwork query error")
    def test_peplot_no_shared_limits(self):
        """
        Test image similarity with no shared limits on the peplot
        """
        visualizer = PredictionError(Ridge(random_state=43), shared_limits=False)

        visualizer.fit(self.data.X.train, self.data.y.train)
        visualizer.score(self.data.X.test, self.data.y.test)
        visualizer.finalize()

        xlim = tuple(map(int, visualizer.ax.get_xlim()))
        ylim = tuple(map(int, visualizer.ax.get_ylim()))
        assert not xlim == ylim

        self.assert_images_similar(visualizer, tol=1.0, remove_legend=True)

    def test_peplot_no_lines(self):
        """
        Test image similarity with no lines drawn on the plot
        """
        visualizer = PredictionError(
            Lasso(random_state=23, alpha=10), bestfit=False, identity=False
        )

        visualizer.fit(self.data.X.train, self.data.y.train)
        visualizer.score(self.data.X.test, self.data.y.test)
        visualizer.finalize()

        self.assert_images_similar(visualizer, tol=1.0, remove_legend=True)

    def test_alpha_param(self):
        """
        Test that the user can supply an alpha param on instantiation
        """
        # Instantiate a sklearn regressor
        model = Lasso(random_state=23, alpha=10)
        # Instantiate a prediction error plot, provide custom alpha
        visualizer = PredictionError(model, bestfit=False, identity=False, alpha=0.7)

        # Test param gets set correctly
        assert visualizer.alpha == 0.7

        # Mock ax and fit the visualizer
        visualizer.ax = mock.MagicMock(autospec=True)
        visualizer.fit(self.data.X.train, self.data.y.train)
        visualizer.score(self.data.X.test, self.data.y.test)

        # Test that alpha was passed to internal matplotlib scatterplot
        _, scatter_kwargs = visualizer.ax.scatter.call_args
        assert "alpha" in scatter_kwargs
        assert scatter_kwargs["alpha"] == 0.7

    def test_is_fitted_param(self):
        """
        Test that the user can supply an is_fitted param and it's state is maintained
        """
        # Instantiate a sklearn regressor
        model = Lasso(random_state=23, alpha=10)
        # Instantiate a prediction error plot, provide custom alpha
        visualizer = PredictionError(
            model, bestfit=False, identity=False, is_fitted=False
        )

        # Test param gets set correctly
        assert visualizer.is_fitted is False

    @pytest.mark.xfail(
        reason="""third test fails with AssertionError: Expected fit
        to be called once. Called 0 times."""
    )
    def test_peplot_with_fitted(self):
        """
        Test that PredictionError properly handles an already-fitted model
        """
        X, y = load_energy(return_dataset=True).to_numpy()

        model = Ridge().fit(X, y)

        with mock.patch.object(model, "fit") as mockfit:
            oz = PredictionError(model)
            oz.fit(X, y)
            mockfit.assert_not_called()

        with mock.patch.object(model, "fit") as mockfit:
            oz = PredictionError(model, is_fitted=True)
            oz.fit(X, y)
            mockfit.assert_not_called()

        with mock.patch.object(model, "fit") as mockfit:
            oz = PredictionError(model, is_fitted=False)
            oz.fit(X, y)
            mockfit.assert_called_once_with(X, y)

    @pytest.mark.xfail(
        IS_WINDOWS_OR_CONDA,
        reason="font rendering different in OS and/or Python; see #892",
    )
    def test_prediction_error_quick_method(self):
        """
        Image similarity test using the residuals plot quick method
        """
        _, ax = plt.subplots()

        model = Lasso(random_state=19)
        oz = prediction_error(
            model, self.data.X.train, self.data.y.train, ax=ax, show=False
        )
        assert isinstance(oz, PredictionError)
        self.assert_images_similar(oz, tol=0.013)

    def test_within_pipeline(self):
        """
        Test that visualizer can be accessed within a sklearn pipeline
        """
        X, y = load_concrete()
        X_train, X_test, y_train, y_test = tts(X, y, test_size=0.2, random_state=42)

        model = Pipeline(
            [
                ("imputer", SimpleImputer(missing_values=np.nan, strategy="mean")),
                ("pe", PredictionError(Lasso())),
            ]
        )

        model.fit(X_train, y_train)
        model.score(X_test, y_test)
        model["pe"].finalize()
        self.assert_images_similar(model["pe"], tol=2.0)

    def test_within_pipeline_quickmethod(self):
        """
        Test that visualizer can be accessed within a sklearn pipeline
        """
        X, y = load_concrete()
        X_train, X_test, y_train, y_test = tts(X, y, test_size=0.2, random_state=42)

        model = Pipeline(
            [
                ("imputer", SimpleImputer(missing_values=np.nan, strategy="mean")),
                ("pe", PredictionError(Lasso())),
            ]
        )

        model.fit(X_train, y_train)
        model.score(X_test, y_test)
        model["pe"].finalize()
        self.assert_images_similar(model["pe"], tol=2.0)

    def test_pipeline_as_model_input(self):
        """
        Test that visualizer can handle sklearn pipeline as model input
        """
        X, y = load_concrete()
        X_train, X_test, y_train, y_test = tts(X, y, test_size=0.2, random_state=42)

        model = Pipeline(
            [
                ("imputer", SimpleImputer(missing_values=np.nan, strategy="mean")),
                ("lasso", Lasso()),
            ]
        )

        oz = PredictionError(model)
        oz.fit(X_train, y_train)
        oz.score(X_test, y_test)
        oz.finalize()
        self.assert_images_similar(oz, tol=2.0)

    def test_pipeline_as_model_input_quickmethod(self):
        """
        Test that visualizer can handle sklearn pipeline as model input
        """
        X, y = load_concrete()
        X_train, X_test, y_train, y_test = tts(X, y, test_size=0.2, random_state=42)

        model = Pipeline(
            [
                ("imputer", SimpleImputer(missing_values=np.nan, strategy="mean")),
                ("lasso", Lasso()),
            ]
        )

        oz = prediction_error(model, X_train, y_train, X_test, y_test)
        oz.finalize()
        self.assert_images_similar(oz, tol=2.0)



# tests.test_datasets.test_signature
# Test the sha256sum file signature library
#
# Author:  Benjamin Bengfort <benjamin@bengfort.com>
# Created: Tue Jul 31 14:20:10 2018 -0400
#
# ID: test_signature.py [7082742] benjamin@bengfort.com $

"""
Test the sha256sum file signature library
"""

##########################################################################
## Imports
##########################################################################

import json
from yellowbrick.datasets.signature import sha256sum


##########################################################################
## Test Case
##########################################################################

FIXTURE = {
    "name": "The Cat in the Hat",
    "color": "red and black",
    "weather": "rainy",
    "chaos_level": "HIGH",
    "things": ["1", "2"],
    "extra": "angry fish in bowl",
}


def test_signature(tmpdir):
    """
    Test the SHA 256 signature of a temporary file
    """

    fpath = tmpdir.join("test.json")
    json.dump(FIXTURE, fpath, indent=2)
    assert (
        sha256sum(str(fpath))
        == "d10b36aa74a59bcf4a88185837f658afaf3646eff2bb16c3928d0e9335e945d2"
    )
